"""The picket fence module is meant for analyzing EPID images where a "picket fence" MLC pattern has been made.
Physicists regularly check MLC positioning through this test. This test can be done using film and one can
"eyeball" it, but this is the 21st century and we have numerous ways of quantifying such data. This module
attains to be one of them. It can load in an EPID dicom image (or superimpose multiple images) and determine the MLC peaks, error of each MLC
pair to the picket, and give a few visual indicators for passing/warning/failing.

Features:

* **Analyze any MLC type** - Both default MLCs and custom MLCs can be used.
* **Easy-to-read pass/warn/fail overlay** - Analysis gives you easy-to-read tools for determining the status of an MLC pair.
* **Any Source-to-Image distance** - Whatever your clinic uses as the SID for picket fence, pylinac can account for it.
* **Account for panel translation** - Have an off-CAX setup? No problem. Translate your EPID and pylinac knows.
* **Account for panel sag** - If your EPID sags at certain angles, just tell pylinac and the results will be shifted.
"""
import copy
import dataclasses
import enum
import io
import os.path as osp
import textwrap
import warnings
import webbrowser
from dataclasses import dataclass
from io import BytesIO
from itertools import cycle
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Union, Tuple, List, Optional, BinaryIO, Iterable, Sequence

import matplotlib.pyplot as plt
import numpy as np
from cached_property import cached_property
from mpl_toolkits.axes_grid1 import make_axes_locatable
from py_linq import Enumerable

from .core import image, pdf
from .core.geometry import Line, Rectangle, Point
from .core.io import get_url, retrieve_demo_file
from .core.profile import MultiProfile, SingleProfile, Interpolation
from .core.utilities import ResultBase, convert_to_enum
from .log_analyzer import load_log
from .settings import get_dicom_cmap

LEFT_MLC_PREFIX = "A"
RIGHT_MLC_PREFIX = "B"


class Orientation(enum.Enum):
    """Possible orientations of the image"""

    UP_DOWN = "Up-Down"  #:
    LEFT_RIGHT = "Left-Right"  #:


class MLCArrangement:
    """Construct an MLC array"""

    def __init__(self, leaf_arrangement: List[Tuple[int, float]], offset: float = 0):
        """

        Parameters
        ----------
        leaf_arrangement: Description of the leaf arrangement. List of tuples containing the number of leaves and leaf width.
            E.g. (10, 5) is 10 leaves with 5mm widths.
        offset: The offset in mm of the leaves. Used for asymmetric arrangements. E.g. -2.5mm will shift the arrangement 2.5mm to the left.
        """
        self.centers = []
        self.widths = []
        rolling_edge = 0
        for leaf_num, width in leaf_arrangement:
            self.centers += np.arange(
                start=rolling_edge + width / 2,
                stop=leaf_num * width + rolling_edge + width / 2,
                step=width,
            ).tolist()
            rolling_edge = self.centers[-1] + width / 2
            self.widths += [width] * leaf_num
        self.centers = [c - np.mean(self.centers) + offset for c in self.centers]


class MLC(enum.Enum):
    """The pre-built MLC types"""

    MILLENNIUM = {
        "name": "Millennium",
        "arrangement": MLCArrangement([(10, 10), (40, 5), (10, 10)]),
    }  #:
    HD_MILLENNIUM = {
        "name": "HD Millennium",
        "arrangement": MLCArrangement([(10, 5), (40, 2.5), (10, 5)]),
    }  #:
    HALCYON_DISTAL = {
        "name": "Halcyon distal",
        "arrangement": MLCArrangement([(60, 5)], offset=2.5),
    }  #:
    HALCYON_PROXIMAL = {
        "name": "Halcyon proximal",
        "arrangement": MLCArrangement([(60, 5)]),
    }  #:


@dataclass
class PFResult(ResultBase):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    tolerance_mm: float  #:
    action_tolerance_mm: float  #:
    percent_leaves_passing: float  #:
    number_of_pickets: int  #:
    absolute_median_error_mm: float  #:
    max_error_mm: float  #:
    mean_picket_spacing_mm: float  #:
    offsets_from_cax_mm: List[float]  #:
    passed: bool  #:
    failed_leaves: Union[List[str], List[int]]  #:
    mlc_skew: float  #:


class PFDicomImage(image.LinacDicomImage):
    """A subclass of a DICOM image that checks for noise and inversion when instantiated. Can also adjust for EPID sag."""

    def __init__(self, path: str, **kwargs):
        crop_mm = kwargs.pop("crop_mm", 3)
        super().__init__(path, **kwargs)
        # crop the images so that Elekta images don't fail. See #168
        crop_pixels = int(round(crop_mm * self.dpmm))
        self.crop(pixels=crop_pixels)
        # self.invert()  # EPID images are always inverted; rather than check inversion, just flip it.
        self._check_for_noise()
        # Possibly revert/change the below if inversion detection doesn't work so well
        self.check_inversion(box_size=10, position=(0.01, 0.01))

    def _check_for_noise(self) -> None:
        """Check if the image has extreme noise (dead pixel, etc) by comparing
        min/max to 1/99 percentiles and smoothing if need be."""
        safety_stop = 5
        while self._has_noise() and safety_stop > 0:
            self.filter(size=3)
            safety_stop -= 1

    def _has_noise(self) -> bool:
        """Helper method to determine if there is spurious signal in the image."""
        min = self.array.min()
        max = self.array.max()
        near_min, near_max = np.percentile(self.array, [0.5, 99.5])
        max_is_extreme = max > near_max * 1.25
        min_is_extreme = (min < near_min * 0.75) and (
            abs(min - near_min) > 0.1 * (near_max - near_min)
        )
        return max_is_extreme or min_is_extreme

    def adjust_for_sag(self, sag: int, orientation: Union[str, Orientation]) -> None:
        """Roll the image to adjust for EPID sag."""
        orient = convert_to_enum(orientation, Orientation)
        direction = "y" if orient == Orientation.UP_DOWN else "x"
        self.roll(direction, sag)


class PicketFence:
    """A class used for analyzing EPID images where radiation strips have been formed by the
    MLCs. The strips are assumed to be parallel to one another and normal to the image edge;
    i.e. a "left-right" or "up-down" orientation is assumed. Further work could follow up by accounting
    for any angle.
    """

    def __init__(
        self,
        filename: Union[str, Path, BinaryIO],
        filter: Optional[int] = None,
        log: Optional[str] = None,
        use_filename: bool = False,
        mlc: Union[MLC, MLCArrangement, str] = MLC.MILLENNIUM,
        crop_mm: int = 3,
        image_kwargs: Optional[dict] = None,
    ):
        """
        Parameters
        ----------
        filename
            Name of the file as a string or a file-like object.
        filter
            If None (default), no filtering will be done to the image.
            If an int, will perform median filtering over image of size ``filter``.
        log
            Path to a log file corresponding to the delivery. The expected fluence of the log file is
            used to construct the pickets. MLC peaks are then compared to an absolute reference instead of
            a fitted picket.
        use_filename
            If False (default), no action will be performed.
            If True, the filename will be searched for keywords that describe the gantry and/or collimator angle.
            For example, if set to True and the file name was "PF_gantry45.dcm" the gantry would be interpreted as being at 45 degrees.
        mlc
            The MLC model of the image. Must be an option from the enum :class:`~pylinac.picketfence.MLCs` or
            an :class:`~pylinac.picketfence.MLCArrangement`.
        crop_mm
            The number of mm to crop from all edges. Elekta is infamous for having columns of dead pixels on the side of their images.
            These need to be cleaned up first. For Varian images, this really shouldn't make a difference unless the pickets are
            very close to the edge. Generally speaking, they shouldn't be for the best accuracy.
        """
        leaf_analysis_width: float  #:
        mlc_meas: List  #:
        pickets: List  #:
        tolerance: float  #:
        action_tolerance: float  #:
        image: PFDicomImage  #:

        if filename is not None:
            img_kwargs = image_kwargs or {}
            self.image = PFDicomImage(
                filename, use_filenames=use_filename, crop_mm=crop_mm, **img_kwargs
            )
            if isinstance(filter, int):
                self.image.filter(size=filter)
            self.image.ground()
            self.image.normalize()
        if log is not None:
            self._load_log(log)
        else:
            self._log_fits = None
        self._is_analyzed = False
        self.mlc = self._get_mlc_arrangement(mlc)

    @staticmethod
    def _get_mlc_arrangement(value: Union[MLC, MLCArrangement, str]) -> MLCArrangement:
        if isinstance(value, MLC):
            return value.value["arrangement"]
        if isinstance(value, MLCArrangement):
            return value
        if isinstance(value, str):
            return [
                member.value["arrangement"]
                for name, member in MLC.__members__.items()
                if member.value["name"] == value
            ][0]

    @classmethod
    def from_url(
        cls, url: str, filter: int = None, image_kwargs: Optional[dict] = None
    ):
        """Instantiate from a URL."""
        filename = get_url(url, progress_bar=True)
        return cls(filename, filter=filter, image_kwargs=image_kwargs)

    @classmethod
    def from_demo_image(cls, filter: int = None):
        """Construct a PicketFence instance using the demo image."""
        demo_file = retrieve_demo_file(name="AS1200.dcm")
        return cls(demo_file, filter=filter)

    @classmethod
    def from_multiple_images(
        cls,
        path_list: Iterable[Union[str, Path]],
        dtype: np.dtype = np.uint16,
        **kwargs,
    ):
        """Load and superimpose multiple images and instantiate a Starshot object.

        Parameters
        ----------
        path_list : iterable
            An iterable of path locations to the files to be loaded/combined.
        kwargs
            Passed to :func:`~pylinac.core.image.load_multiples`.
        """
        obj = cls.from_demo_image()
        # save a combined image to a temporary dir, then load it back in as a PFDicomImage
        with TemporaryDirectory() as tmp:
            filename = osp.join(tmp, "mydcm.dcm")
            image.load_multiples(path_list, dtype=dtype, **kwargs).save(filename)
            obj.image = PFDicomImage(filename)
        return obj

    @property
    def passed(self) -> bool:
        """Boolean specifying if all MLC positions were within tolerance."""
        # nested all because each measurement returns a list of booleans. So all the passes of each measurement must all pass.
        return all(all(m.passed) for m in self.mlc_meas)

    @property
    def percent_passing(self) -> float:
        """Return the percentage of MLC positions under tolerance."""
        num_meas = Enumerable(self.mlc_meas).select_many(lambda m: m.passed).count()
        num_pass = (
            Enumerable(self.mlc_meas)
            .select_many(lambda m: m.passed)
            .count(lambda p: p == True)
        )
        return float(100 * num_pass / num_meas)

    @property
    def max_error(self) -> float:
        """Return the maximum error found."""
        return float(np.max(np.abs(self._flattened_errors())))

    @property
    def max_error_picket(self) -> int:
        """Return the picket number where the maximum error occurred."""
        return (
            Enumerable(self.mlc_meas)
            .order_by_descending(lambda m: np.max(np.abs(m.error)))
            .select(lambda m: m.picket_num)
            .first()
        )

    @property
    def max_error_leaf(self) -> int:
        """Return the leaf that had the maximum error."""
        if not self.separate_leaves:
            return (
                Enumerable(self.mlc_meas)
                .order_by_descending(lambda m: np.max(np.abs(m.error)))
                .select(lambda m: m.full_leaf_nums)
                .first()
            )
        else:
            max_meas = (
                Enumerable(self.mlc_meas)
                .order_by_descending(lambda m: np.max(np.abs(m.error)))
                .first()
            )
            if abs(max_meas.error[0]) > abs(max_meas.error[1]):
                return max_meas.full_leaf_nums[0]
            else:
                return max_meas.full_leaf_nums[1]

    def _flattened_errors(self) -> List[float]:
        return Enumerable(self.mlc_meas).select_many(lambda m: m.error).to_list()

    def failed_leaves(self) -> Union[List[int], List[str]]:
        """A list of the failed leaves. Either the leaf number or the bank+leaf number if using separate leaves."""
        if not self._is_analyzed:
            raise ValueError(
                "It appears the PF image has not been analyzed yet. Use .analyze() first."
            )
        failing_sets = Enumerable(self.mlc_meas).where(lambda m: not all(m.passed))
        if not self.separate_leaves:
            return failing_sets.select(lambda m: m.leaf_num).distinct().to_list()
        else:
            return (
                failing_sets.select_many(
                    lambda m: [
                        m.full_leaf_nums[idx]
                        for idx, passed in enumerate(m.passed)
                        if not passed
                    ]
                )
                .distinct()
                .to_list()
            )

    @property
    def abs_median_error(self) -> float:
        """Return the median error found."""
        return float(np.median(np.abs(self._flattened_errors())))

    @property
    def num_pickets(self) -> int:
        """Return the number of pickets determined."""
        return len(self.pickets)

    @property
    def mean_picket_spacing(self) -> float:
        """The average distance between pickets in mm."""
        sorted_pickets = sorted(self.pickets, key=lambda x: x.dist2cax)
        return float(
            np.mean(
                [
                    abs(sorted_pickets[idx].dist2cax - sorted_pickets[idx + 1].dist2cax)
                    for idx in range(len(sorted_pickets) - 1)
                ]
            )
        )

    def plot_leaf_profile(self, leaf: Union[str, int], picket: int, show: bool = True):
        """Plot the leaf profile of a given leaf pair parallel to leaf motion.

        Parameters
        ----------
        leaf
            The leaf to plot. If ``separate_leaves`` is True, this will be a string like "A15" or "B33".
            If ``separate_leaves`` is False, this must be an int, like ``15`` or ``33``.
        picket
            An int of the picket number. Pickets start from the 0-side of an image. E.g. for left-right PFs, this would start on the left; for up-down this would start at the bottom.
        """
        mlc_meas = Enumerable(self.mlc_meas).single(
            lambda m: leaf in m.full_leaf_nums and m.picket_num == picket
        )
        ax = mlc_meas.plot_detailed_profile()
        ax.set_title(f"MLC profile Leaf: {leaf}, Picket: {picket}")
        for lg, rg, m in zip(
            self.pickets[picket].left_guard_separated,
            self.pickets[picket].right_guard_separated,
            mlc_meas.marker_lines,
        ):
            g_val = lg(m.point1.y)
            rg_val = rg(m.point1.y)
            ax.axvline(g_val, color="green", label="Guard rail")
            ax.axvline(rg_val, color="green", label="Guard rail")
        ax.legend()
        if show:
            plt.show()

    def save_leaf_profile(
        self,
        filename: Union[str, Path, BinaryIO],
        leaf: Union[str, int],
        picket: int,
        **kwargs,
    ):
        """Save the leaf profile plot to disk or stream. See plot_leaf_profile for parameter hints. Kwargs are passed to matplotlib.savefig()"""
        self.plot_leaf_profile(leaf, picket, show=False)
        plt.savefig(filename, **kwargs)
        if not isinstance(filename, BytesIO):
            print(f"Picket fence leaf profile saved to: {osp.abspath(filename)}")

    def _load_log(self, log: str) -> None:
        """Load a machine log that corresponds to the picket fence delivery.

        This log determines the location of the pickets. The MLC peaks are then compared to the expected log pickets,
        not a simple fit of the peaks."""
        # load the log fluence image
        mlog = load_log(log)
        fl = mlog.fluence.expected.calc_map(equal_aspect=True)
        fli = image.load(
            fl, dpi=254
        )  # 254 pix/in => 1 pix/0.1mm (default fluence calc)

        # equate them such that they're the same size & DPI
        fluence_img, img_array = image.equate_images(fli, self.image)
        self.image.array = img_array.array

        # get picket fits from the modified fluence image
        pf = PicketFence.from_demo_image()
        pf.image = fluence_img
        pf.analyze()
        self._log_fits = cycle([p.get_fit() for p in pf.pickets])

    @staticmethod
    def run_demo(tolerance: float = 0.5, action_tolerance: float = None) -> None:
        """Run the Picket Fence demo using the demo image. See analyze() for parameter info."""
        pf = PicketFence.from_demo_image()
        pf.analyze(tolerance, action_tolerance=action_tolerance)
        print(pf.results())
        pf.plot_analyzed_image(leaf_error_subplot=True)

    def analyze(
        self,
        tolerance: float = 0.5,
        action_tolerance: Optional[float] = None,
        num_pickets: Optional[int] = None,
        sag_adjustment: Union[float, int] = 0,
        orientation: Optional[Union[Orientation, str]] = None,
        invert: bool = False,
        leaf_analysis_width_ratio: float = 0.4,
        picket_spacing: Optional[float] = None,
        height_threshold: float = 0.5,
        edge_threshold: float = 1.5,
        peak_sort: str = "peak_heights",
        required_prominence: float = 0.2,
        fwxm: int = 50,
        separate_leaves: bool = False,
        nominal_gap_mm: float = 3,
    ) -> None:
        """Analyze the picket fence image.

        Parameters
        ----------
        tolerance
            The tolerance of difference in mm between an MLC pair position and the
            picket fit line.
        action_tolerance
            If None (default), no action tolerance is set or compared to.
            If an int or float, the MLC pair measurement is also compared to this
            tolerance. Must be lower than tolerance. This value is usually meant
            to indicate that a physicist should take an "action" to reduce the error,
            but should not stop treatment.
        num_pickets
            The number of pickets in the image. A helper parameter to limit the total number of pickets,
            only needed if analysis is catching more pickets than there really are.
        sag_adjustment
            The amount of shift in mm to apply to the image to correct for EPID sag.
            For Up-Down picket images, positive moves the image down, negative up.
            For Left-Right picket images, positive moves the image left, negative right.
        orientation
            If None (default), the orientation is automatically determined. If for some reason the determined
            orientation is not correct, you can pass it directly using this parameter.
            If passed a string with 'u' (e.g. 'up-down', 'u-d', 'up') it will set the orientation of the pickets as
            going up-down. If passed a string with 'l' (e.g. 'left-right', 'lr', 'left') it will set it as going
            left-right.
        invert
            If False (default), the inversion of the image is automatically detected and used.
            If True, the image inversion is reversed from the automatic detection. This is useful when runtime errors
            are encountered.
        leaf_analysis_width_ratio
            The ratio of the leaf width to use as part of the evaluation. E.g. if the ratio is 0.5, the center half of
            the leaf will be used. This helps avoid tongue and groove influence.
        picket_spacing
            If None (default), the spacing between pickets is determined automatically.
            If given, it should be an int or float specifying the number of **PIXELS** apart the pickets are.
        height_threshold
            The threshold that the MLC peak needs to be above to be considered a picket (vs background).
            Lower if not all leaves are being caught. Note that for FFF beams this would very likely need to be lowered.
        edge_threshold
            The threshold of pixel value standard deviation within the analysis window of the MLC leaf to be considered a full leaf.
            This is how pylinac removes MLCs that are eclipsed by the jaw. This also is how to
            omit or catch leaves at the edge of the field. Raise to catch more edge leaves.
        peak_sort
            Either 'peak_heights' or 'prominences'. This is the method for determining the peaks. Usually not needed
            unless the wrong number of pickets have been detected.
            See the scipy.signal.find_peaks function for more information.
        required_prominence
            The required height of the picket (not individual MLCs) to be considered a peak.
            Pylinac takes a mean of the image axis perpendicular to the leaf motion to get an initial guess of the peak
            locations and also to determine picket spacing. Changing this can be useful for wide-gap tests where
            the shape of the beam horns can form two or more local maximums in the picket area. Increase if for wide-gap
            images that are catching too many pickets. Consider lowering for FFF beams if there are analysis issues.

            .. warning::

                We do not recommend performing FFF wide-gap PF tests. Make your FFF pickets narrow or measure with a flat beam instead.

        fwxm
            For each MLC kiss, the profile is a curve from low to high to low. The FWXM (0-100) is the height to use to measure
            to determine the center of the curve, which is the surrogate for MLC kiss position. I.e. for each MLC kiss,
            what height of the picket should you use to actually determine the center location? It is unusual to change this.
            If you have something in the way (we've seen crazy examples with a BB in the way) you may want to increase this.
        separate_leaves
            Whether to analyze leaves individually (each tip) or as a set (combined, center of the picket). False is
            the default for backwards compatibility.
        nominal_gap_mm
            The expected gap of the pickets in mm. Only used when separate leaves is True. Due to the DLG and EPID
            scattering, this value will have to be determined by you with a known good delivery.
        """
        if action_tolerance is not None and tolerance < action_tolerance:
            raise ValueError("Tolerance cannot be lower than the action tolerance")
        self.tolerance = tolerance
        self.action_tolerance = action_tolerance
        self.leaf_analysis_width = leaf_analysis_width_ratio
        self.separate_leaves = separate_leaves

        if invert:
            self.image.invert()

        self._orientation = orientation
        # adjust for sag
        if sag_adjustment != 0:
            sag_pixels = int(round(sag_adjustment * self.image.dpmm))
            self.image.adjust_for_sag(sag_pixels, self.orientation)

        if self.orientation == Orientation.UP_DOWN:
            leaf_prof = np.mean(self.image, 0)
        else:
            leaf_prof = np.mean(self.image, 1)
        leaf_prof = MultiProfile(leaf_prof)
        peak_idxs, peak_vals = leaf_prof.find_fwxm_peaks(
            min_distance=0.02,
            threshold=height_threshold,
            max_number=num_pickets,
            peak_sort=peak_sort,
            required_prominence=required_prominence,
        )
        if len(peak_idxs) == 0:
            raise ValueError(
                "No pickets were found. This can mean either an incorrect orientation or incorrect inversion. "
                "Try passing the correct orientation; if that fails, also set invert=True."
            )
        # get picket spacing if not set by user
        if picket_spacing is None:
            picket_spacing = np.median(np.diff(np.sort(peak_idxs)))

        # loop through each leaf row and analyze each MLC kiss
        self.mlc_meas = []
        for leaf_num, center, width in self._leaves_in_view(leaf_analysis_width_ratio):
            for picket_num, (picket_idx, picket_peak_val) in enumerate(
                zip(peak_idxs, peak_vals)
            ):
                window = self._get_mlc_window(
                    leaf_center=center,
                    leaf_width=width,
                    approx_idx=picket_idx,
                    spacing=picket_spacing,
                )
                if self._is_mlc_peak_in_window(
                    window, height_threshold, edge_threshold, picket_peak_val
                ):
                    self.mlc_meas.append(
                        MLCValue(
                            picket_num=picket_num,
                            approx_idx=picket_idx,
                            leaf_width=width,
                            leaf_center=center,
                            picket_spacing=picket_spacing,
                            orientation=self.orientation,
                            leaf_analysis_width_ratio=leaf_analysis_width_ratio,
                            tolerance=tolerance,
                            action_tolerance=action_tolerance,
                            leaf_num=leaf_num,
                            approx_peak_val=picket_peak_val,
                            image_window=window,
                            image=self.image,
                            fwxm=fwxm,
                            separate_leaves=separate_leaves,
                            nominal_gap_mm=nominal_gap_mm,
                        )
                    )
        if not self.mlc_meas:
            raise ValueError(
                "No MLC measurements were found. This may be due to an incorrect inversion. Try setting invert=True. Or, you may have passed an incorrect orientation."
            )

        # drop any leaf rows that don't have the right amount of MLC kisses (i.e. near edge where one is dropped)
        median_num_leaves = (
            Enumerable(self.mlc_meas)
            .group_by(key=lambda m: m.leaf_num)
            .median(lambda m: len(m))
        )
        full_leaves = (
            Enumerable(self.mlc_meas)
            .group_by(key=lambda m: m.leaf_num)
            .where(lambda m: len(m) == median_num_leaves)
            .select_many(lambda m: m)
            .select(lambda m: m.leaf_num)
            .distinct()
            .to_list()
        )
        self.mlc_meas = [m for m in self.mlc_meas if m.leaf_num in full_leaves]
        if any([True for m in self.mlc_meas if m.leaf_num not in full_leaves]):
            warnings.warn(
                "Some leaves were removed from analysis because they were not detected for all pickets. If some valid leaves are missing try adjusting height_threshold or edge_threshold"
            )

        # retrospectively create the pickets and update the individual MLC measurements so error can be calculated
        self.pickets = []
        for picket_num, _ in enumerate(peak_idxs):
            self.pickets.append(
                Picket(
                    [m for m in self.mlc_meas if m.picket_num == picket_num],
                    log_fits=self._log_fits,
                    orientation=self.orientation,
                    image=self.image,
                    tolerance=tolerance,
                    nominal_gap=nominal_gap_mm,
                    separate_leaves=separate_leaves,
                )
            )

        self._is_analyzed = True

    def _is_mlc_peak_in_window(
        self, window, height_threshold, edge_threshold, picket_peak_val
    ) -> bool:
        """Whether the MLC peak is inside the given window. E.g. the jaw could be closed or at an edge."""
        if self.orientation == Orientation.UP_DOWN:
            std = np.std(window, axis=1)
        else:
            std = np.std(window, axis=0)
        is_above_height_threshold = np.max(window) > height_threshold * picket_peak_val
        is_not_at_edge = max(std) < edge_threshold * np.median(std)
        return is_above_height_threshold and is_not_at_edge

    def _get_mlc_window(
        self, leaf_center, leaf_width, approx_idx, spacing
    ) -> np.ndarray:
        """A small 2D window of the image that contains the area around the picket."""
        leaf_width_px = leaf_width * self.image.dpmm
        leaf_center_px = leaf_center * self.image.dpmm + (
            self.image.shape[0] / 2
            if self.orientation == Orientation.UP_DOWN
            else self.image.shape[1] / 2
        )
        if self.orientation == Orientation.UP_DOWN:
            # crop edges to image boundary if need be; if the pickets are too close to edge we could spill outside
            left_edge = max(int(approx_idx - spacing / 2), 0)
            right_edge = min(int(approx_idx + spacing / 2), self.image.shape[1])
            top_edge = max(int(leaf_center_px - leaf_width_px / 2), 0)
            bottom_edge = min(
                int(leaf_center_px + leaf_width_px / 2), self.image.shape[0]
            )
            array = self.image[top_edge:bottom_edge, left_edge:right_edge]
        else:
            top_edge = max(int(approx_idx - spacing / 2), 0)
            bottom_edge = min(int(approx_idx + spacing / 2), self.image.shape[0])
            left_edge = max(int(leaf_center_px - leaf_width_px / 2), 0)
            right_edge = min(
                int(leaf_center_px + leaf_width_px / 2), self.image.shape[1]
            )
            array = self.image[top_edge:bottom_edge, left_edge:right_edge]
        return array

    def _leaves_in_view(self, analysis_width) -> List[Tuple[int, int, int]]:
        """Crop the leaves if not all leaves are in view."""
        range = (
            self.image.shape[0] / 2
            if self.orientation == Orientation.UP_DOWN
            else self.image.shape[1] / 2
        )
        # cut off the edge so that we're not halfway through a leaf.
        range -= (
            max(
                self.mlc.widths[0] * analysis_width,
                self.mlc.widths[-1] * analysis_width,
            )
            * self.image.dpmm
        )
        leaves = [
            i
            for i, c in enumerate(self.mlc.centers)
            if abs(c) < (range / self.image.dpmm)
        ]
        return [
            (leaf_num, center, width)
            for leaf_num, center, width in zip(
                leaves,
                self.mlc.centers[leaves[0] : leaves[-1] + 1],
                self.mlc.widths[leaves[0] : leaves[-1] + 1],
            )
        ]

    def plot_analyzed_image(
        self,
        guard_rails: bool = True,
        mlc_peaks: bool = True,
        overlay: bool = True,
        leaf_error_subplot: bool = True,
        show: bool = True,
        figure_size: Union[str, Tuple] = "auto",
    ) -> None:
        """Plot the analyzed image.

        Parameters
        ----------
        guard_rails
            Do/don't plot the picket "guard rails" around the ideal picket
        mlc_peaks
            Do/don't plot the detected MLC peak positions.
        overlay
            Do/don't plot the alpha overlay of the leaf status.
        leaf_error_subplot
            If True, plots a linked leaf error subplot adjacent to the PF image plotting the average and standard
            deviation of leaf error.
        show
            Whether to display the plot. Set to false for saving to a figure, etc.
        figure_size
            Either 'auto' or a tuple. If auto, the figure size is set depending on the orientation. If a tuple, this is the
            figure size to use.
        """
        if not self._is_analyzed:
            raise RuntimeError("The image must be analyzed first. Use .analyze().")

        # plot the image
        if figure_size == "auto":
            if self.orientation == Orientation.UP_DOWN:
                figure_size = (12, 8)
            else:
                figure_size = (9, 9)
        fig, ax = plt.subplots(figsize=figure_size)
        ax.imshow(self.image.array, cmap=get_dicom_cmap())

        if leaf_error_subplot:
            self._add_leaf_error_subplot(ax)

        if guard_rails:
            for picket in self.pickets:
                picket.add_guards_to_axes(ax.axes)
        if mlc_peaks:
            for mlc_meas in self.mlc_meas:
                mlc_meas.plot2axes(ax.axes, width=1.5)

        if overlay:
            for mlc_meas in self.mlc_meas:
                mlc_meas.plot_overlay2axes(ax.axes)

        # plot CAX
        ax.plot(
            self.image.center.x, self.image.center.y, "r+", ms=12, markeredgewidth=3
        )

        # tighten up the plot view
        ax.set_xlim([0, self.image.shape[1]])
        ax.set_ylim([0, self.image.shape[0]])
        ax.axis("off")

        if show:
            plt.show()

    def _add_leaf_error_subplot(self, ax: plt.Axes) -> None:
        """Add a bar subplot showing the leaf error."""

        # make the new axis
        divider = make_axes_locatable(ax)
        if self.orientation == Orientation.UP_DOWN:
            axtop = divider.append_axes("right", 2, pad=1, sharey=ax)
        else:
            axtop = divider.append_axes("bottom", 2, pad=1, sharex=ax)

        # get leaf positions, errors, standard deviation, and leaf numbers
        if self.orientation == Orientation.UP_DOWN:
            pos = [
                position.marker_lines[0].center.y
                for position in self.pickets[0].mlc_meas
            ]
        else:
            pos = [
                position.marker_lines[0].center.x
                for position in self.pickets[0].mlc_meas
            ]

        # calculate the error and stdev values per MLC pair
        error_stdev = []
        error_vals = []
        for leaf_num in set([m.leaf_num for m in self.mlc_meas]):
            error_vals.append(
                np.mean(
                    [np.abs(m.error) for m in self.mlc_meas if m.leaf_num == leaf_num]
                )
            )
            error_stdev.append(
                np.std([m.error for m in self.mlc_meas if m.leaf_num == leaf_num])
            )

        # plot the leaf errors as a bar plot
        if self.orientation == Orientation.UP_DOWN:
            axtop.barh(
                pos,
                error_vals,
                xerr=error_stdev,
                height=self.leaf_analysis_width * 2,
                alpha=0.4,
                align="center",
            )
            # plot the tolerance line(s)
            axtop.axvline(self.tolerance, color="r", linewidth=3)
            if self.action_tolerance is not None:
                axtop.axvline(self.action_tolerance, color="m", linewidth=3)
            # reset xlims to comfortably include the max error or tolerance value
            axtop.set_xlim(
                [0, max([max(error_vals) + max(error_stdev), self.tolerance]) + 0.1]
            )
        else:
            axtop.bar(
                pos,
                error_vals,
                yerr=error_stdev,
                width=self.leaf_analysis_width * 2,
                alpha=0.4,
                align="center",
            )
            # plot the tolerance line(s)
            axtop.axhline(self.tolerance, color="r", linewidth=3)
            if self.action_tolerance is not None:
                axtop.axhline(self.action_tolerance, color="m", linewidth=3)
            axtop.set_ylim(
                [0, max([max(error_vals) + max(error_stdev), self.tolerance]) + 0.1]
            )

        axtop.grid(True)
        axtop.set_title("Average Error (mm)")

    def save_analyzed_image(
        self,
        filename: Union[str, io.BytesIO],
        guard_rails: bool = True,
        mlc_peaks: bool = True,
        overlay: bool = True,
        leaf_error_subplot: bool = False,
        **kwargs,
    ) -> None:
        """Save the analyzed figure to a file. See :meth:`~pylinac.picketfence.PicketFence.plot_analyzed_image()` for
        further parameter info.
        """
        self.plot_analyzed_image(
            guard_rails,
            mlc_peaks,
            overlay,
            leaf_error_subplot=leaf_error_subplot,
            show=False,
        )
        plt.savefig(filename, **kwargs)
        if isinstance(filename, str):
            print(f"Picket fence image saved to: {osp.abspath(filename)}")

    def results(self) -> str:
        """Return results of analysis. Use with print()."""
        pass_pct = self.percent_passing
        offsets = " ".join("{:.1f}".format(pk.dist2cax) for pk in self.pickets)
        string = (
            f"Picket Fence Results: \n{pass_pct:2.1f}% "
            f"Passed\nMedian Error: {self.abs_median_error:2.3f}mm \n"
            f"Mean picket spacing: {self.mean_picket_spacing:2.1f}mm \n"
            f"Picket offsets from CAX (mm): {offsets}\n"
            f"Max Error: {self.max_error:2.3f}mm on Picket: {self.max_error_picket}, Leaf: {self.max_error_leaf}\n"
            f"Failing leaves: {self.failed_leaves()}"
        )
        return string

    def results_data(self, as_dict=False) -> Union[PFResult, dict]:
        """Present the results data and metadata as a dataclass, dict, or tuple.
        The default return type is a dataclass."""
        data = PFResult(
            tolerance_mm=self.tolerance,
            action_tolerance_mm=self.action_tolerance,
            percent_leaves_passing=self.percent_passing,
            number_of_pickets=self.num_pickets,
            absolute_median_error_mm=self.abs_median_error,
            max_error_mm=self.max_error,
            mean_picket_spacing_mm=self.mean_picket_spacing,
            offsets_from_cax_mm=[pk.dist2cax for pk in self.pickets],
            passed=self.passed,
            failed_leaves=self.failed_leaves(),
            mlc_skew=self.mlc_skew(),
        )
        if as_dict:
            return dataclasses.asdict(data)
        return data

    def publish_pdf(
        self,
        filename: Union[str, io.BytesIO],
        notes: str = None,
        open_file: bool = False,
        metadata: dict = None,
        bins: int = 10,
    ) -> None:
        """Publish (print) a PDF containing the analysis, images, and quantitative results.

        Parameters
        ----------
        filename : (str, file-like object}
            The file to write the results to.
        notes : str, list of strings
            Text; if str, prints single line.
            If list of strings, each list item is printed on its own line.
        open_file : bool
            Whether to open the file using the default program after creation.
        metadata : dict
            Extra data to be passed and shown in the PDF. The key and value will be shown with a colon.
            E.g. passing {'Author': 'James', 'Unit': 'TrueBeam'} would result in text in the PDF like:
            --------------
            Author: James
            Unit: TrueBeam
            --------------
        """
        plt.ioff()
        canvas = pdf.PylinacCanvas(
            filename, page_title="Picket Fence Analysis", metadata=metadata
        )
        data = io.BytesIO()
        self.save_analyzed_image(data, leaf_error_subplot=True)
        canvas.add_image(data, location=(3, 8), dimensions=(15, 15))
        text = [
            "Picket Fence results:",
            f"Magnification factor (SID/SAD): {self.image.metadata.RTImageSID / self.image.metadata.RadiationMachineSAD:2.2f}",
            f"Tolerance (mm): {self.tolerance}",
            f"Leaves passing (%): {self.percent_passing:2.1f}",
            f"Absolute median error (mm): {self.abs_median_error:2.3f}",
            f"Mean picket spacing (mm): {self.mean_picket_spacing:2.1f}",
            f"Maximum error (mm): {self.max_error:2.3f} on Picket {self.max_error_picket}, Leaf {self.max_error_leaf}",
        ]
        if self.failed_leaves():
            text += textwrap.wrap(f"Failing leaves: {self.failed_leaves()}")
        text.append(f"Gantry Angle: {self.image.gantry_angle:2.2f}")
        text.append(f"Collimator Angle: {self.image.collimator_angle:2.2f}")
        canvas.add_text(text=text, location=(10, 25.5))
        if notes is not None:
            canvas.add_text(text="Notes:", location=(1, 5.5), font_size=14)
            canvas.add_text(text=notes, location=(1, 5))

        canvas.add_new_page()
        hist = io.BytesIO()
        self.save_histogram(hist, bins)
        canvas.add_image(hist, location=(3, 8), dimensions=(15, 15))
        canvas.finish()

        if open_file:
            webbrowser.open(filename)

    def mlc_skew(self) -> float:
        """Apparent rotation in degrees of the MLC. This could be conflated with the EPID skew, so be careful when interpreting this value."""
        return Enumerable(self.pickets).avg(lambda p: np.rad2deg(np.arctan(p.fit.c[0])))

    def plot_histogram(self, bins: int = 10, show: bool = True) -> None:
        """Plot a histogram of the leaf errors"""
        if not self._is_analyzed:
            raise ValueError(
                "It appears the PF image has not been analyzed yet. Use .analyze() first."
            )
        errors = Enumerable(self.mlc_meas).select_many(lambda m: m.error).to_list()
        fig, ax = plt.subplots()
        ax.set_title("Leaf error histogram")
        ax.set_ylabel("Counts")
        ax.set_xlabel("Error (mm)")
        ax.hist(errors, bins=bins)
        if show:
            plt.show()

    def save_histogram(
        self, filename: [str, Path, BinaryIO], bins: int = 10, **kwargs
    ) -> None:
        """Save a histogram of the leaf errors"""
        self.plot_histogram(bins, show=False)
        plt.savefig(filename, **kwargs)
        if not isinstance(filename, BytesIO):
            print(f"Picket fence histogram saved to: {osp.abspath(filename)}")

    @cached_property
    def orientation(self) -> Orientation:
        """The orientation of the image, either Up-Down or Left-Right."""
        # if orientation was passed in, use it
        if self._orientation is not None:
            return convert_to_enum(self._orientation, Orientation)

        # replace any dead pixels with median value
        temp_image = self.image.array.copy()
        temp_image[temp_image < np.median(temp_image)] = np.median(temp_image)

        # find "range" of 80 to 90th percentiles
        row_sum = np.sum(temp_image, 0)
        col_sum = np.sum(temp_image, 1)
        row80, row90 = np.percentile(row_sum, [85, 99])
        col80, col90 = np.percentile(col_sum, [85, 99])
        row_range = row90 - row80
        col_range = col90 - col80

        # The true picket side will have a greater difference in
        # percentiles than will the non-picket size.
        if row_range < col_range:
            orientation = Orientation.LEFT_RIGHT
        else:
            orientation = Orientation.UP_DOWN
        return orientation


class MLCValue:
    def __init__(
        self,
        picket_num: int,
        approx_idx: int,
        leaf_width: float,
        leaf_center: float,
        picket_spacing: float,
        orientation: Orientation,
        leaf_analysis_width_ratio: float,
        tolerance: float,
        action_tolerance: Optional[float],
        leaf_num: int,
        approx_peak_val: float,
        image_window: np.ndarray,
        image: PFDicomImage,
        fwxm: int,
        separate_leaves: bool,
        nominal_gap_mm: float,
    ):
        """Representation of an MLC kiss or of each MLC about a kiss."""
        self._approximate_idx = approx_idx
        self.picket_num = picket_num
        self._approximate_peak_vale = approx_peak_val
        self.leaf_width_px = leaf_width * image.dpmm
        self._leaf_center = leaf_center
        self.leaf_center_px = leaf_center * image.dpmm + (
            image.shape[0] / 2
            if orientation == Orientation.UP_DOWN
            else image.shape[1] / 2
        )
        self.leaf_num = leaf_num
        self._image_window = image_window
        self._image = image
        self._fwxm = fwxm
        self._analysis_ratio: float = leaf_analysis_width_ratio
        self._spacing: float = picket_spacing
        self._orientation = orientation
        self._tolerance: float = tolerance
        self._action_tolerance: float = action_tolerance
        self._separate_leaves = separate_leaves
        self._nominal_gap_mm = nominal_gap_mm
        self.profile: SingleProfile
        self.position = self.get_peak_positions()
        self._fit = None

    def __repr__(self) -> str:
        return f"Leaf: {self.leaf_num}, Picket: {self.picket_num}"

    @property
    def full_leaf_nums(self) -> Sequence[Union[str, int]]:
        """The fully-qualified leaf names. This will be the simple leaf number for traditional analysis or the bank+leaf num for separate leaves."""
        if not self._separate_leaves:
            return [
                self.leaf_num,
            ]
        else:
            return [
                f"{LEFT_MLC_PREFIX}{self.leaf_num}",
                f"{RIGHT_MLC_PREFIX}{self.leaf_num}",
            ]

    def plot2axes(self, axes: plt.Axes, width: Union[float, int] = 1) -> None:
        """Plot the measurement to the axes."""
        for idx, line in enumerate(self.marker_lines):
            line.plot2axes(axes, width, color=self.bg_color[idx])

    def get_peak_positions(self) -> Sequence[float]:
        if self._orientation == Orientation.UP_DOWN:
            pix_vals = np.median(self._image_window, axis=0)
        else:
            pix_vals = np.median(self._image_window, axis=1)
        interpolation_factor = 100
        prof = SingleProfile(
            pix_vals,
            interpolation=Interpolation.LINEAR,
            interpolation_factor=interpolation_factor,
        )
        fwxm = prof.fwxm_data(self._fwxm)
        self.profile = prof
        if self._separate_leaves:
            left = fwxm["left index (exact)"] + max(
                self._approximate_idx - self._spacing / 2, 0
            )
            right = fwxm["right index (exact)"] + max(
                self._approximate_idx - self._spacing / 2, 0
            )
            return left, right
        else:
            return (
                fwxm["center index (exact)"]
                + max(self._approximate_idx - self._spacing / 2, 0),
            )  # crop to left edge if need be

    @property
    def passed(self) -> Sequence[bool]:
        """Whether the MLC kiss or leaf was within tolerance."""
        return [abs(error) < self._tolerance for error in self.error]

    @property
    def passed_action(self) -> Optional[Sequence[bool]]:
        """Whether the MLC kiss or leaf was within the action tolerance."""
        return (
            [abs(error) < self._action_tolerance for error in self.error]
            if self._action_tolerance is not None
            else [True, True]
        )

    @property
    def bg_color(self) -> Sequence[str]:
        """The color of the measurement when the PF image is plotted, based on pass/fail status."""
        colors = []
        for idx, passed in enumerate(self.passed):
            if not passed:
                colors.append("r")
            elif self._action_tolerance is not None:
                colors.append("b" if self.passed_action[idx] else "m")
            else:
                colors.append("b")
        return colors

    @property
    def picket_positions(self) -> Sequence[float]:
        """The position(s) of the pickets in mm"""
        picket_pos = []
        for line, sign in zip(self.marker_lines, (-1, 1)):
            if self._orientation == Orientation.UP_DOWN:
                picket = self._fit(line.center.y)
            else:
                picket = self._fit(line.center.x)
            if (
                self._separate_leaves
            ):  # offset the picket position by the DLG and nominal gap
                mag_factor = self._image.sid / 1000
                picket += (
                    sign * self._nominal_gap_mm * mag_factor / 2 * self._image.dpmm
                )
            picket_pos.append(picket / self._image.dpmm)
        return picket_pos

    def plot_detailed_profile(self) -> plt.Axes:
        if self._orientation == Orientation.UP_DOWN:
            pix_vals = np.median(self._image_window, axis=0)
        else:
            pix_vals = np.median(self._image_window, axis=1)
        offset_pixels = max(self._approximate_idx - self._spacing / 2, 0)
        x_values = np.array(range(len(pix_vals))) + offset_pixels

        fig, ax = plt.subplots()
        ax.plot(x_values, pix_vals)
        # ax.vlines(x=data['center index (exact)']/self.profile._interpolation_factor + offset_pixels, ymin=data['peak_props']['peak_heights'][0] - data['peak_props']['prominences'][0], ymax=data['peak_props']['peak_heights'][0], color='magenta', label='Peak center')
        for picket_pos in self.picket_positions:
            ax.axvline(
                x=picket_pos * self._image.dpmm,
                label="Fitted picket location",
                color="black",
            )
        for pos, bg_color in zip(self.get_peak_positions(), self.bg_color):
            ax.axvline(pos, color=bg_color, label="Measured MLC position")
        return ax

    @property
    def error(self) -> Sequence[float]:
        """The error (difference) of the MLC measurement and the picket fit.
        If using individual leaf analysis, returns both errors otherwise return one."""
        errors = []
        for line, sign in zip(self.marker_lines, (-1, 1)):
            if self._orientation == Orientation.UP_DOWN:
                picket_pos = self._fit(line.center.y)
                mlc_pos = line.center.x
            else:
                picket_pos = self._fit(line.center.x)
                mlc_pos = line.center.y
            if (
                self._separate_leaves
            ):  # offset the picket position by the DLG and nominal gap
                mag_factor = self._image.sid / 1000
                picket_pos += (
                    sign * self._nominal_gap_mm * mag_factor / 2 * self._image.dpmm
                )
            errors.append((mlc_pos - picket_pos) / self._image.dpmm)
        return errors

    @property
    def max_abs_error(self) -> float:
        """The maximum absolute error"""
        return np.max(np.abs([self.error]))

    @property
    def marker_lines(self) -> List[Line]:
        """The line(s) representing the MLC measurement position. When using separated leaves
        there are two lines. Traditional analysis returns one."""
        upper_point = (
            self.leaf_center_px - self.leaf_width_px / 2 * self._analysis_ratio
        )
        lower_point = (
            self.leaf_center_px + self.leaf_width_px / 2 * self._analysis_ratio
        )

        lines = []
        for mlc_position in self.position:
            if self._orientation == Orientation.UP_DOWN:
                line = Line((mlc_position, upper_point), (mlc_position, lower_point))
            else:
                line = Line((upper_point, mlc_position), (lower_point, mlc_position))
            lines.append(line)
        return lines

    def plot_overlay2axes(self, axes) -> None:
        """Create a rectangle overlay with the width of the error. I.e. it stretches from the picket fit to the MLC position. Gives more visual size to the"""
        # calculate height (based on leaf analysis ratio)
        upper_point = (
            self.leaf_center_px - self.leaf_width_px / 2 * self._analysis_ratio
        )
        lower_point = (
            self.leaf_center_px + self.leaf_width_px / 2 * self._analysis_ratio
        )
        height = abs(upper_point - lower_point) * 0.8

        for idx, line in enumerate(self.marker_lines):
            width = abs(self.error[idx]) * self._image.dpmm
            y = line.center.y
            x = self.position[idx] - (self.error[idx] * self._image.dpmm) / 2

            if self._orientation == Orientation.UP_DOWN:
                r = Rectangle(width, height, center=(x, y))
                # if any of the values are over tolerance, show another larger rectangle to draw the eye
                if not self.passed[idx] or not self.passed_action[idx]:
                    re = Rectangle(
                        self._image_window.shape[1] * 0.2, height * 1.2, center=(x, y)
                    )
                    re.plot2axes(
                        axes,
                        edgecolor="none",
                        fill=True,
                        alpha=0.5,
                        facecolor=self.bg_color[idx],
                    )
            else:
                r = Rectangle(height, width, center=(x, y))
                if not self.passed[idx] or not self.passed_action[idx]:
                    re = Rectangle(
                        self._image_window.shape[1] * 0.2, height * 1.2, center=(x, y)
                    )
                    re.plot2axes(
                        axes,
                        edgecolor="none",
                        fill=True,
                        alpha=0.5,
                        facecolor=self.bg_color[idx],
                    )
            r.plot2axes(
                axes, edgecolor="none", fill=True, alpha=1, facecolor=self.bg_color[idx]
            )


class Picket:
    """Holds picket information in a Picket Fence test."""

    def __init__(
        self,
        mlc_measurements: List[MLCValue],
        log_fits,
        orientation: Orientation,
        image: PFDicomImage,
        tolerance: float,
        separate_leaves: bool,
        nominal_gap: float,
    ):
        self.mlc_meas: List[MLCValue] = mlc_measurements
        self.log_fits = log_fits
        self.tolerance = tolerance
        self.orientation = orientation
        self.image = image
        self._separate_leaves = separate_leaves
        self._nominal_gap = nominal_gap
        self.fit = self.get_fit()
        for m in self.mlc_meas:
            m._fit = self.fit

    def get_fit(self) -> np.poly1d:
        """The fit of a polynomial to the MLC measurements."""
        if self.log_fits is not None:
            return next(self.log_fits)
        x = (
            Enumerable(self.mlc_meas)
            .select_many(lambda m: [l.point1.y for l in m.marker_lines])
            .to_list()
        )
        y = (
            Enumerable(self.mlc_meas)
            .select_many(lambda m: [l.point1.x for l in m.marker_lines])
            .to_list()
        )
        if self.orientation == Orientation.UP_DOWN:
            fit = np.polyfit(x, y, 1)
        else:
            fit = np.polyfit(y, x, 1)
        return np.poly1d(fit)

    @property
    def dist2cax(self) -> float:
        """The distance from the CAX to the picket, in mm."""
        # TODO: see about using line and built-in dist to point.
        center_fit = np.poly1d(self.fit)
        if self.orientation == Orientation.UP_DOWN:
            length = self.image.shape[0]
        else:
            length = self.image.shape[1]
        x_data = np.arange(length)
        y_data = center_fit(x_data)
        idx = int(round(len(x_data) / 2))
        if self.orientation == Orientation.UP_DOWN:
            axis = "x"
            p1 = Point(y_data[idx], x_data[idx])
        else:
            axis = "y"
            p1 = Point(x_data[idx], y_data[idx])
        return (getattr(self.image.center, axis) - getattr(p1, axis)) / self.image.dpmm

    @property
    def left_guard_separated(self) -> Sequence[np.poly1d]:
        """The line representing the left-sided guard rails.
        When not doing separate analysis, the left and right rails will overlap."""
        l_fit = np.copy(self.fit)
        l_fit[-1] += self.tolerance * self.image.dpmm
        if not self._separate_leaves:
            return [
                np.poly1d(l_fit),
            ]
        else:
            other_fit = copy.copy(l_fit)
            mag_factor = self.image.sid / 1000
            l_fit[-1] += self._nominal_gap * mag_factor / 2 * self.image.dpmm
            other_fit[-1] -= self._nominal_gap * mag_factor / 2 * self.image.dpmm
            return [np.poly1d(l_fit), np.poly1d(other_fit)]

    @property
    def right_guard_separated(self):
        """The line representing the right-sided guard rails."""
        r_fit = np.copy(self.fit)
        r_fit[-1] -= self.tolerance * self.image.dpmm
        if not self._separate_leaves:
            return [
                np.poly1d(r_fit),
            ]
        else:
            other_fit = copy.copy(r_fit)
            mag_factor = self.image.sid / 1000
            r_fit[-1] -= self._nominal_gap * mag_factor / 2 * self.image.dpmm
            other_fit[-1] += self._nominal_gap * mag_factor / 2 * self.image.dpmm
            return [np.poly1d(r_fit), np.poly1d(other_fit)]

    def add_guards_to_axes(self, axis: plt.Axes, color: str = "g") -> None:
        """Plot guard rails to the axis."""
        if self.orientation == Orientation.UP_DOWN:
            length = self.image.shape[0]
        else:
            length = self.image.shape[1]
        x_data = np.arange(length)
        left_y_data = self.left_guard_separated
        right_y_data = self.right_guard_separated
        for left, right in zip(left_y_data, right_y_data):
            if self.orientation == Orientation.UP_DOWN:
                axis.plot(left(x_data), x_data, color=color)
                axis.plot(right(x_data), x_data, color=color)
            else:
                axis.plot(x_data, left(x_data), color=color)
                axis.plot(x_data, right(x_data), color=color)



"""The planar imaging module analyzes phantom images taken with the kV or MV imager in 2D.
The following phantoms are supported:

* Leeds TOR 18
* Standard Imaging QC-3
* Standard Imaging QC-kV
* Las Vegas
* Doselab MC2 MV
* Doselab MC2 kV
* SNC kV
* SNC MV
* PTW EPID QC

Features:

* **Automatic phantom localization** - Set up your phantom any way you like; automatic positioning,
  angle, and inversion correction mean you can set up how you like, nor will setup variations give you headache.
* **High and low contrast determination** - Analyze both low and high contrast ROIs. Set thresholds
  as you see fit.
"""
import copy
import dataclasses
import io
import math
import os.path as osp
import warnings
import webbrowser
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, List, Tuple, Union, BinaryIO, Callable, Dict

import matplotlib.pyplot as plt
import numpy as np
from cached_property import cached_property
from skimage import feature, measure
from skimage.measure._regionprops import RegionProperties

from .core import image, pdf, geometry
from .core.decorators import lru_cache
from .core.geometry import Point, Rectangle, Circle, Vector
from .core.io import get_url, retrieve_demo_file
from .core.mtf import MTF
from .core.profile import Interpolation, CollapsedCircleProfile, SingleProfile
from .core.roi import LowContrastDiskROI, HighContrastDiskROI, bbox_center, Contrast
from .core.utilities import ResultBase
from .ct import get_regions


@dataclass
class PlanarResult(ResultBase):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    analysis_type: str  #:
    median_contrast: float  #:
    median_cnr: float  #:
    num_contrast_rois_seen: int  #:
    phantom_center_x_y: Tuple[float, float]  #:
    mtf_lp_mm: Tuple[float, float, float] = None  #:


def _middle_of_bbox_region(region: RegionProperties) -> Tuple:
    return (
        (region.bbox[2] - region.bbox[0]) / 2 + region.bbox[0],
        (region.bbox[3] - region.bbox[1]) / 2 + region.bbox[1],
    )


def is_centered(region: RegionProperties, instance: object, rtol=0.3) -> bool:
    """Whether the region is centered on the image"""
    img_center = (instance.image.center.y, instance.image.center.x)
    # we don't want centroid because that could be offset by missing lengths of the outline. Center of bbox is more robust
    return np.allclose(_middle_of_bbox_region(region), img_center, rtol=rtol)


def is_right_size(region: RegionProperties, instance: object, rtol=0.1) -> bool:
    """Whether the region is close to the expected size of the phantom, given the SSD and physical phantom size."""
    return bool(
        np.isclose(
            region.bbox_area,
            instance.phantom_bbox_size_px / (instance._ssd / 1000) ** 2,
            rtol=rtol,
        )
    )


class ImagePhantomBase:
    """Base class for planar phantom classes.

    Attributes
    ----------
    common_name : str
        The human-readable name of the phantom. Used in plots and PDF report.
    phantom_outline_object : {None, 'Circle', 'Rectangle'}
        What type of outline to display on the plotted image. Helps to visually determine the accuracy of the
        phantom size, position, and scale.
    high_contrast_rois : list
        :class:`~pylinac.core.roi.HighContrastDiskROI` instances of the
        high contrast line pair regions.
    high_contrast_roi_settings : dict
        Settings of the placement of the high-contrast ROIs.
    low_contrast_rois : list
        :class:`~pylinac.core.roi.LowContrastDiskROI` instances of the low
        contrast ROIs, other than the reference ROI (below).
    low_contrast_roi_settings : dict
        Settings of the placement of the low-contrast ROIs.
    low_contrast_background_rois : list
        :class:`~pylinac.core.roi.LowContrastDiskROI` instances of the low
        contrast background ROIs.
    low_contrast_background_roi_settings : dict
        Settings of the placement of the background low-contrast ROIs.
    low_contrast_background_value : float
        The average pixel value of all the low-contrast background ROIs.
    detection_conditions: list of callables
        This should be a list of functions that return a boolean. It is used for finding the phantom outline in the image.
        E.g. is_at_center().
    phantom_bbox_size_mm2: float
        This is the expected size of the **BOUNDING BOX** of the phantom. Additionally, it is usually smaller than the
        physical bounding box because we sometimes detect an inner ring/square. Typically, x0.9-1.0 of the physical size.
    """

    _demo_filename: str
    common_name: str
    high_contrast_roi_settings = {}
    high_contrast_rois = []
    low_contrast_roi_settings = {}
    low_contrast_rois = []
    low_contrast_background_roi_settings = {}
    low_contrast_background_rois = []
    low_contrast_background_value = None
    phantom_outline_object = None
    detection_conditions: [List[Callable]] = [is_centered, is_right_size]
    detection_canny_settings = {"sigma": 2, "percentiles": (0.001, 0.01)}
    phantom_bbox_size_mm2: float

    def __init__(
        self,
        filepath: Union[str, BinaryIO, Path],
        normalize: bool = True,
        image_kwargs: Optional[dict] = None,
    ):
        """
        Parameters
        ----------
        filepath : str
            Path to the image file.
        normalize: bool
            Whether to "ground" and normalize the image. This can affect contrast measurements, but for
            backwards compatibility this is True. You may want to set this to False if trying to compare with other software.
        image_kwargs : dict
            Keywords passed to the image load function; this would include things like DPI or SID if applicable
        """
        img_kwargs = image_kwargs or {}
        self.image = image.load(filepath, **img_kwargs)
        if normalize:
            self.image.ground()
            self.image.normalize()
        self._angle_override = None
        self._size_override = None
        self._center_override = None
        self._high_contrast_threshold = None
        self._low_contrast_threshold = None
        self._ssd: float = 100
        self.mtf = None

    @classmethod
    def from_demo_image(cls):
        """Instantiate and load the demo image."""
        demo_file = retrieve_demo_file(name=cls._demo_filename)
        return cls(demo_file)

    @classmethod
    def from_url(cls, url: str):
        """
        Parameters
        ----------
        url : str
            The URL to the image.
        """
        image_file = get_url(url)
        return cls(image_file)

    def _preprocess(self):
        pass

    def _check_inversion(self):
        pass

    @property
    def phantom_bbox_size_px(self) -> float:
        """The phantom bounding box size in pixels^2."""
        return self.phantom_bbox_size_mm2 * (self.image.dpmm**2)

    @cached_property
    def phantom_ski_region(self) -> RegionProperties:
        """The skimage region of the phantom outline."""
        regions = self._get_canny_regions()
        # search through all the canny ROIs to see which ones pass the detection conditions
        blobs = []
        for phantom_idx, region in enumerate(regions):
            conditions_met = [
                condition(region, self) for condition in self.detection_conditions
            ]
            if all(conditions_met):
                blobs.append(phantom_idx)

        if not blobs:
            raise ValueError(
                "Unable to find the phantom in the image. Potential solutions: check the SSD was passed correctly, check that the phantom isn't at the edge of the field, check that the phantom is centered along the CAX."
            )

        # take the biggest ROI and call that the phantom outline
        big_roi_idx = np.argsort([regions[phan].area_bbox for phan in blobs])[-1]
        phantom_idx = blobs[big_roi_idx]

        return regions[phantom_idx]

    def analyze(
        self,
        low_contrast_threshold: float = 0.05,
        high_contrast_threshold: float = 0.5,
        invert: bool = False,
        angle_override: Optional[float] = None,
        center_override: Optional[tuple] = None,
        size_override: Optional[float] = None,
        ssd: float = 1000,
        low_contrast_method: Contrast = Contrast.MICHELSON,
        visibility_threshold: float = 100,
    ) -> None:
        """Analyze the phantom using the provided thresholds and settings.

        Parameters
        ----------
        low_contrast_threshold : float
            This is the contrast threshold value which defines any low-contrast ROI as passing or failing.
        high_contrast_threshold : float
            This is the contrast threshold value which defines any high-contrast ROI as passing or failing.
        invert : bool
            Whether to force an inversion of the image. This is useful if pylinac's automatic inversion algorithm fails
            to properly invert the image.
        angle_override : None, float
            A manual override of the angle of the phantom. If None, pylinac will automatically determine the angle. If
            a value is passed, this value will override the automatic detection.

            .. Note::

                0 is pointing from the center toward the right and positive values go counterclockwise.

        center_override : None, 2-element tuple
            A manual override of the center point of the phantom. If None, pylinac will automatically determine the center. If
            a value is passed, this value will override the automatic detection. Format is (x, y)/(col, row).
        size_override : None, float
            A manual override of the relative size of the phantom. This size value is used to scale the positions of
            the ROIs from the center. If None, pylinac will automatically determine the size.
            If a value is passed, this value will override the automatic sizing.

            .. Note::

                 This value is not necessarily the physical size of the phantom. It is an arbitrary value.
        ssd
            The SSD of the phantom itself in mm.
        low_contrast_method
            The equation to use for calculating low contrast.
        visibility_threshold
            The threshold for whether an ROI is "seen".
        """
        self._angle_override = angle_override
        self._center_override = center_override
        self._size_override = size_override
        self._high_contrast_threshold = high_contrast_threshold
        self._low_contrast_threshold = low_contrast_threshold
        self._low_contrast_method = low_contrast_method
        self.visibility_threshold = visibility_threshold
        self._ssd = ssd
        self._check_inversion()
        if invert:
            self.image.invert()
        self._preprocess()
        if self.high_contrast_roi_settings:
            self.high_contrast_rois = self._sample_high_contrast_rois()
            # generate rMTF
            spacings = [
                roi["lp/mm"] for roi in self.high_contrast_roi_settings.values()
            ]
            self.mtf = MTF.from_high_contrast_diskset(
                diskset=self.high_contrast_rois, spacings=spacings
            )
        if self.low_contrast_background_roi_settings:
            (
                self.low_contrast_background_rois,
                self.low_contrast_background_value,
            ) = self._sample_low_contrast_background_rois()
        if self.low_contrast_roi_settings:
            self.low_contrast_rois = self._sample_low_contrast_rois()

    def _sample_low_contrast_rois(self) -> List[LowContrastDiskROI]:
        """Sample the low-contrast sample regions for calculating contrast values."""
        lc_rois = []
        for stng in self.low_contrast_roi_settings.values():
            roi = LowContrastDiskROI(
                self.image,
                self.phantom_angle + stng["angle"],
                self.phantom_radius * stng["roi radius"],
                self.phantom_radius * stng["distance from center"],
                self.phantom_center,
                self._low_contrast_threshold,
                self.low_contrast_background_value,
                contrast_method=self._low_contrast_method,
                visibility_threshold=self.visibility_threshold,
            )
            lc_rois.append(roi)
        return lc_rois

    def _sample_low_contrast_background_rois(
        self,
    ) -> Tuple[List[LowContrastDiskROI], float]:
        """Sample the low-contrast background regions for calculating contrast values."""
        bg_rois = []
        for stng in self.low_contrast_background_roi_settings.values():
            roi = LowContrastDiskROI(
                self.image,
                self.phantom_angle + stng["angle"],
                self.phantom_radius * stng["roi radius"],
                self.phantom_radius * stng["distance from center"],
                self.phantom_center,
                self._low_contrast_threshold,
            )
            bg_rois.append(roi)
        avg_bg = np.mean([roi.pixel_value for roi in bg_rois])
        return bg_rois, avg_bg

    def _sample_high_contrast_rois(self) -> List[HighContrastDiskROI]:
        """Sample the high-contrast line pair regions."""
        hc_rois = []
        for stng in self.high_contrast_roi_settings.values():
            roi = HighContrastDiskROI(
                self.image,
                self.phantom_angle + stng["angle"],
                self.phantom_radius * stng["roi radius"],
                self.phantom_radius * stng["distance from center"],
                self.phantom_center,
                self._high_contrast_threshold,
            )
            hc_rois.append(roi)
        return hc_rois

    def save_analyzed_image(
        self,
        filename: Union[None, str, BinaryIO] = None,
        split_plots: bool = False,
        to_streams: bool = False,
        **kwargs,
    ) -> Optional[Union[Dict[str, BinaryIO], List[str]]]:
        """Save the analyzed image to disk or to stream. Kwargs are passed to plt.savefig()

        Parameters
        ----------
        filename : None, str, stream
            A string representing where to save the file to. If split_plots and to_streams are both true, leave as None as newly-created streams are returned.
        split_plots: bool
            If split_plots is True, multiple files will be created that append a name. E.g. `my_file.png` will become `my_file_image.png`, `my_file_mtf.png`, etc.
            If to_streams is False, a list of new filenames will be returned
        to_streams: bool
            This only matters if split_plots is True. If both of these are true, multiple streams will be created and returned as a dict.
        """
        if filename is None and to_streams is False:
            raise ValueError("Must pass in a filename unless saving to streams.")
        figs, names = self.plot_analyzed_image(
            show=False, split_plots=split_plots, **kwargs
        )
        # remove plot keywords as savefig complains about extra kwargs
        for key in (
            "image",
            "low_contrast",
            "high_contrast",
            "show",
        ):
            kwargs.pop(key, None)
        if not split_plots:
            plt.savefig(filename, **kwargs)
        else:
            # append names to filename if it's file-like
            if not to_streams:
                filenames = []
                f, ext = osp.splitext(filename)
                for name in names:
                    filenames.append(f + "_" + name + ext)
            else:  # it's a stream buffer
                filenames = [io.BytesIO() for _ in names]
            for fig, name in zip(figs, filenames):
                fig.savefig(name, **kwargs)
            if to_streams:
                return {name: stream for name, stream in zip(names, filenames)}
            if split_plots:
                return filenames

    def _get_canny_regions(self) -> List[RegionProperties]:
        """Compute the canny edges of the image and return the connected regions found."""
        # compute the canny edges with very low thresholds (detects nearly everything)
        canny_img = feature.canny(
            self.image.array,
            low_threshold=self.detection_canny_settings["percentiles"][0],
            high_threshold=self.detection_canny_settings["percentiles"][1],
            use_quantiles=True,
            sigma=self.detection_canny_settings["sigma"],
        )

        # label the canny edge regions
        labeled = measure.label(canny_img)
        regions = measure.regionprops(labeled, intensity_image=self.image.array)
        return regions

    def _create_phantom_outline_object(self) -> Tuple[Union[Rectangle, Circle], dict]:
        """Construct the phantom outline object which will be plotted on the image for visual inspection."""
        outline_type = list(self.phantom_outline_object)[0]
        outline_settings = list(self.phantom_outline_object.values())[0]
        settings = {}
        if outline_type == "Rectangle":
            side_a = self.phantom_radius * outline_settings["width ratio"]
            side_b = self.phantom_radius * outline_settings["height ratio"]
            half_hyp = np.sqrt(side_a**2 + side_b**2) / 2
            internal_angle = ia = np.rad2deg(np.arctan(side_b / side_a))
            new_x = self.phantom_center.x + half_hyp * (
                geometry.cos(ia) - geometry.cos(ia + self.phantom_angle)
            )
            new_y = self.phantom_center.y + half_hyp * (
                geometry.sin(ia) - geometry.sin(ia + self.phantom_angle)
            )
            obj = Rectangle(
                width=self.phantom_radius * outline_settings["width ratio"],
                height=self.phantom_radius * outline_settings["height ratio"],
                center=Point(new_x, new_y),
            )
            settings["angle"] = self.phantom_angle
        elif outline_type == "Circle":
            obj = Circle(
                center_point=self.phantom_center,
                radius=self.phantom_radius * outline_settings["radius ratio"],
            )
        else:
            raise ValueError(
                "An outline object was passed but was not a Circle or Rectangle."
            )
        return obj, settings

    def plot_analyzed_image(
        self,
        image: bool = True,
        low_contrast: bool = True,
        high_contrast: bool = True,
        show: bool = True,
        split_plots: bool = False,
        **plt_kwargs: dict,
    ) -> Tuple[List[plt.Figure], List[str]]:
        """Plot the analyzed image.

        Parameters
        ----------
        image : bool
            Show the image.
        low_contrast : bool
            Show the low contrast values plot.
        high_contrast : bool
            Show the high contrast values plot.
        show : bool
            Whether to actually show the image when called.
        split_plots : bool
            Whether to split the resulting image into individual plots. Useful for saving images into individual files.
        plt_kwargs : dict
            Keyword args passed to the plt.figure() method. Allows one to set things like figure size.
        """
        plot_low_contrast = low_contrast and any(self.low_contrast_rois)
        plot_high_contrast = high_contrast and any(self.high_contrast_rois)
        num_plots = sum((image, plot_low_contrast, plot_high_contrast))
        if num_plots < 1:
            warnings.warn(
                "Nothing was plotted because either all parameters were false or there were no actual high/low ROIs"
            )
            return
        # set up axes and make axes iterable
        figs = []
        names = []
        if split_plots:
            axes = []
            for n in range(num_plots):
                fig, axis = plt.subplots(1, **plt_kwargs)
                figs.append(fig)
                axes.append(axis)
        else:
            fig, axes = plt.subplots(1, num_plots, **plt_kwargs)
            fig.subplots_adjust(wspace=0.4)
        if num_plots < 2:
            axes = (axes,)
        axes = iter(axes)

        # plot the marked image
        if image:
            img_ax = next(axes)
            names.append("image")
            self.image.plot(ax=img_ax, show=False)
            img_ax.axis("off")
            img_ax.set_title(f"{self.common_name} Phantom Analysis")

            # plot the outline image
            if self.phantom_outline_object is not None:
                outline_obj, settings = self._create_phantom_outline_object()
                outline_obj.plot2axes(img_ax, edgecolor="b", **settings)
            # plot the low contrast background ROIs
            for roi in self.low_contrast_background_rois:
                roi.plot2axes(img_ax, edgecolor="b")
            # plot the low contrast ROIs
            for roi in self.low_contrast_rois:
                roi.plot2axes(img_ax, edgecolor=roi.plot_color)
            # plot the high-contrast ROIs along w/ pass/fail coloration
            if self.high_contrast_rois:
                for (roi, mtf) in zip(
                    self.high_contrast_rois, self.mtf.norm_mtfs.values()
                ):
                    color = "b" if mtf > self._high_contrast_threshold else "r"
                    roi.plot2axes(img_ax, edgecolor=color)

        # plot the low contrast value graph
        if plot_low_contrast:
            lowcon_ax = next(axes)
            names.append("low_contrast")
            self._plot_lowcontrast_graph(lowcon_ax)

        # plot the high contrast MTF graph
        if plot_high_contrast:
            hicon_ax = next(axes)
            names.append("high_contrast")
            self._plot_highcontrast_graph(hicon_ax)

        plt.tight_layout()
        if show:
            plt.show()
        return figs, names

    def _plot_lowcontrast_graph(self, axes: plt.Axes):
        """Plot the low contrast ROIs to an axes."""
        (line1,) = axes.plot(
            [roi.contrast for roi in self.low_contrast_rois],
            marker="o",
            color="m",
            label="Contrast",
        )
        axes.axhline(self._low_contrast_threshold, color="m")
        axes.grid(True)
        axes.set_title("Low-frequency Contrast")
        axes.set_xlabel("ROI #")
        axes.set_ylabel("Contrast")
        axes2 = axes.twinx()
        (line2,) = axes2.plot(
            [roi.contrast_to_noise for roi in self.low_contrast_rois],
            marker="^",
            label="CNR",
        )
        axes2.set_ylabel("CNR")
        axes.legend(handles=[line1, line2])

    def _plot_highcontrast_graph(self, axes: plt.Axes):
        """Plot the high contrast ROIs to an axes."""
        axes.plot(self.mtf.spacings, list(self.mtf.norm_mtfs.values()), marker="*")
        axes.axhline(self._high_contrast_threshold, color="k")
        axes.grid(True)
        axes.set_title("High-frequency rMTF")
        axes.set_xlabel("Line pairs / mm")
        axes.set_ylabel("relative MTF")

    def results(self) -> str:
        """Return the results of the analysis."""
        text = [f"{self.common_name} results:", f"File: {self.image.truncated_path}"]
        if self.low_contrast_rois:
            text += [
                f"Median Contrast: {np.median([roi.contrast for roi in self.low_contrast_rois]):2.2f}",
                f"Median CNR: {np.median([roi.contrast_to_noise for roi in self.low_contrast_rois]):2.1f}",
                f'# Low contrast ROIs "seen": {sum(roi.passed for roi in self.low_contrast_rois):2.0f} of {len(self.low_contrast_rois)}',
            ]
        if self.high_contrast_rois:
            text += [
                f"MTF 80% (lp/mm): {self.mtf.relative_resolution(80):2.2f}",
                f"MTF 50% (lp/mm): {self.mtf.relative_resolution(50):2.2f}",
                f"MTF 30% (lp/mm): {self.mtf.relative_resolution(30):2.2f}",
            ]
        text = "\n".join(text)
        return text

    def results_data(self, as_dict=False) -> Union[PlanarResult, dict]:
        data = PlanarResult(
            analysis_type=self.common_name,
            median_contrast=np.median([roi.contrast for roi in self.low_contrast_rois]),
            median_cnr=np.median(
                [roi.contrast_to_noise for roi in self.low_contrast_rois]
            ),
            num_contrast_rois_seen=sum(roi.passed for roi in self.low_contrast_rois),
            phantom_center_x_y=(self.phantom_center.x, self.phantom_center.y),
        )

        if self.mtf is not None:
            data.mtf_lp_mm = [
                {p: self.mtf.relative_resolution(p)} for p in (80, 50, 30)
            ]
        if as_dict:
            return dataclasses.asdict(data)
        return data

    def publish_pdf(
        self,
        filename: str,
        notes: str = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ):
        """Publish (print) a PDF containing the analysis, images, and quantitative results.

        Parameters
        ----------
        filename : (str, file-like object}
            The file to write the results to.
        notes : str, list of strings
            Text; if str, prints single line.
            If list of strings, each list item is printed on its own line.
        open_file : bool
            Whether to open the file using the default program after creation.
        metadata : dict
            Extra data to be passed and shown in the PDF. The key and value will be shown with a colon.
            E.g. passing {'Author': 'James', 'Unit': 'TrueBeam'} would result in text in the PDF like:
            --------------
            Author: James
            Unit: TrueBeam
            --------------
        """
        canvas = pdf.PylinacCanvas(
            filename,
            page_title=f"{self.common_name} Phantom Analysis",
            metadata=metadata,
        )

        # write the text/numerical values
        text = self.results()
        canvas.add_text(text=text, location=(1.5, 25), font_size=14)
        if notes is not None:
            canvas.add_text(text="Notes:", location=(1, 5.5), font_size=12)
            canvas.add_text(text=notes, location=(1, 5))

        # plot the image
        data = io.BytesIO()
        self.save_analyzed_image(
            data, image=True, low_contrast=False, high_contrast=False
        )
        canvas.add_image(data, location=(1, 3.5), dimensions=(19, 19))
        # plot the high contrast
        if self.high_contrast_rois:
            canvas.add_new_page()
            data = io.BytesIO()
            self.save_analyzed_image(
                data, image=False, low_contrast=False, high_contrast=True
            )
            canvas.add_image(data, location=(1, 7), dimensions=(19, 19))
        # plot the low contrast
        if self.low_contrast_rois:
            canvas.add_new_page()
            data = io.BytesIO()
            self.save_analyzed_image(
                data, image=False, low_contrast=True, high_contrast=False
            )
            canvas.add_image(data, location=(1, 7), dimensions=(19, 19))

        canvas.finish()
        if open_file:
            webbrowser.open(filename)

    @property
    def phantom_center(self) -> Point:
        return (
            Point(self._center_override)
            if self._center_override is not None
            else self._phantom_center_calc()
        )

    @property
    def phantom_radius(self) -> float:
        return (
            self._size_override
            if self._size_override is not None
            else self._phantom_radius_calc()
        )

    @property
    def phantom_angle(self) -> float:
        return (
            self._angle_override
            if self._angle_override is not None
            else self._phantom_angle_calc()
        )

    def _phantom_center_calc(self):
        return bbox_center(self.phantom_ski_region)

    def _phantom_angle_calc(self):
        pass

    def _phantom_radius_calc(self):
        pass


@dataclass
class LightRadResult(ResultBase):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes."""

    field_size_x_mm: float  #:
    field_size_y_mm: float  #:
    field_epid_offset_x_mm: float  #:
    field_epid_offset_y_mm: float  #:
    field_bb_offset_x_mm: float  #:
    field_bb_offset_y_mm: float  #:


class StandardImagingFC2(ImagePhantomBase):
    common_name = "SI FC-2"
    _demo_filename = "fc2.dcm"
    # these positions are the offset in mm from the center of the image to the nominal position of the BBs
    bb_positions_10x10 = {
        "TL": [-40, -40],
        "BL": [-40, 40],
        "TR": [40, -40],
        "BR": [40, 40],
    }
    bb_positions_15x15 = {
        "TL": [-65, -65],
        "BL": [-65, 65],
        "TR": [65, -65],
        "BR": [65, 65],
    }
    bb_sampling_box_size_mm = 10
    field_strip_width_mm = 5

    @staticmethod
    def run_demo() -> None:
        """Run the Standard Imaging FC-2 phantom analysis demonstration."""
        fc2 = StandardImagingFC2.from_demo_image()
        fc2.analyze()
        fc2.plot_analyzed_image()

    def analyze(self, invert: bool = False, fwxm: int = 50) -> None:
        """Analyze the FC-2 phantom to find the BBs and the open field and compare to each other as well as the EPID.

        Parameters
        ----------
        invert : bool
            Whether to force-invert the image from the auto-detected inversion.
        fwxm : int
            The FWXM value to use to detect the field. For flattened fields, the default of 50 should be fine.
            For FFF fields, consider using a lower value such as 25-30.
        """
        self._check_inversion()
        if invert:
            self.image.invert()
        self.bb_center = self._find_overall_bb_centroid(fwxm=fwxm)
        (
            self.field_center,
            self.field_width_x,
            self.field_width_y,
        ) = self._find_field_info(fwxm=fwxm)
        self.epid_center = self.image.center

    def results(self, as_list: bool = False) -> Union[str, list]:
        """Return the results of the analysis."""
        text = [
            f"{self.common_name} results:",
            f"File: {self.image.truncated_path}",
            f"The detected inplane field size was {self.field_width_y:2.1f}mm",
            f"The detected crossplane field size was {self.field_width_x:2.1f}mm",
            f"The inplane field was {self.field_epid_offset_mm.y:2.1f}mm from the EPID CAX",
            f"The crossplane field was {self.field_epid_offset_mm.x:2.1f}mm from the EPID CAX",
            f"The inplane field was {self.field_bb_offset_mm.y:2.1f}mm from the BB inplane center",
            f"The crossplane field was {self.field_bb_offset_mm.x:2.1f}mm from the BB crossplane center",
        ]
        if as_list:
            return text
        else:
            text = "\n".join(text)
            return text

    @property
    def field_epid_offset_mm(self) -> Vector:
        """Field offset from CAX using vector difference"""
        return (
            self.epid_center.as_vector() - self.field_center.as_vector()
        ) / self.image.dpmm

    @property
    def field_bb_offset_mm(self) -> Vector:
        """Field offset from BB centroid using vector difference"""
        return (
            self.bb_center.as_vector() - self.field_center.as_vector()
        ) / self.image.dpmm

    def results_data(self, as_dict: bool = False) -> Union[LightRadResult, dict]:
        """Return the results as a dict or dataclass"""
        data = LightRadResult(
            field_size_x_mm=self.field_width_x,
            field_size_y_mm=self.field_width_y,
            field_epid_offset_x_mm=self.field_epid_offset_mm.x,
            field_epid_offset_y_mm=self.field_epid_offset_mm.y,
            field_bb_offset_x_mm=self.field_bb_offset_mm.x,
            field_bb_offset_y_mm=self.field_bb_offset_mm.y,
        )
        if as_dict:
            return dataclasses.asdict(data)
        return data

    def _check_inversion(self):
        """Perform a normal corner-check inversion. Since these are always 10x10 or 15x15 fields it seems unlikely the corners will be exposed."""
        self.image.check_inversion()

    def _find_field_info(self, fwxm: int) -> (Point, float, float):
        """Determine the center and field widths of the detected field by sampling a strip through the center of the image in inplane and crossplane"""
        sample_width = self.field_strip_width_mm / 2 * self.image.dpmm
        # sample the strip (nominally 5mm) centered about the image center. Average the strip to reduce noise.
        x_bounds = (
            int(self.image.center.x - sample_width),
            int(self.image.center.x + sample_width),
        )
        y_img = np.mean(self.image[:, x_bounds[0] : x_bounds[1]], 1)
        y_prof = SingleProfile(
            y_img, interpolation=Interpolation.NONE, dpmm=self.image.dpmm
        )
        y_bounds = (
            int(self.image.center.y - sample_width),
            int(self.image.center.y + sample_width),
        )
        x_img = np.mean(self.image[y_bounds[0] : y_bounds[1], :], 0)
        x_prof = SingleProfile(
            x_img, interpolation=Interpolation.NONE, dpmm=self.image.dpmm
        )
        x = x_prof.fwxm_data(x=fwxm)["center index (exact)"]
        y = y_prof.fwxm_data(x=fwxm)["center index (exact)"]
        field_width_x = x_prof.fwxm_data(x=fwxm)["width (exact) mm"]
        field_width_y = y_prof.fwxm_data(x=fwxm)["width (exact) mm"]
        return Point(x=x, y=y), field_width_x, field_width_y

    def _find_overall_bb_centroid(self, fwxm: int) -> Point:
        """Determine the geometric center of the 4 BBs"""
        self.bb_centers = bb_centers = self._detect_bb_centers(fwxm)
        central_x = np.mean([p.x for p in bb_centers.values()])
        central_y = np.mean([p.y for p in bb_centers.values()])
        return Point(x=central_x, y=central_y)

    def _detect_bb_centers(self, fwxm: int) -> dict:
        """Sample a 10x10mm square about each BB to detect it. Adjustable using self.bb_sampling_box_size_mm"""
        bb_positions = {}
        nominal_positions = self._determine_bb_set(fwxm=fwxm)
        dpmm = self.image.dpmm
        sample_box_size = self.bb_sampling_box_size_mm / 2 * dpmm
        # invert the image so that the BB marks increase in intensity
        inverted_img = copy.copy(self.image)
        inverted_img.invert()
        # sample the square, use skimage to find the ROI weighted centroid of the BBs
        for key, position in nominal_positions.items():
            x_bounds = (
                int(self.image.center.x + (position[0] * dpmm) - sample_box_size),
                int(self.image.center.x + (position[0] * dpmm) + sample_box_size),
            )
            y_bounds = (
                int(self.image.center.y + (position[1] * dpmm) - sample_box_size),
                int(self.image.center.y + (position[1] * dpmm) + sample_box_size),
            )
            bb_sample = image.load(
                inverted_img[y_bounds[0] : y_bounds[1], x_bounds[0] : x_bounds[1]]
            )
            bb_sample.ground()
            _, rprops, num_roi = get_regions(
                bb_sample.array, fill_holes=True, clear_borders=False
            )
            if num_roi < 1:
                raise ValueError("Did not find the BB in the expected location")
            center_roi = take_centermost_roi(rprops, bb_sample.array.shape)
            # due to the sloping field values, the centroid is usually a better representation.
            # the weighted centroid will bias towards the center of the field. Even though this isn't technically a problem,
            # I know people will complain it's not aligned with the BB.
            bb_positions[key] = Point(
                x=center_roi.centroid[1] + x_bounds[0],
                y=center_roi.centroid[0] + y_bounds[0],
            )
        return bb_positions

    def _determine_bb_set(self, fwxm: int) -> dict:
        """This finds the approximate field size to determine whether to check for the 10x10 BBs or the 15x15. Returns the BB positions"""
        x_prof = SingleProfile(
            self.image[int(self.image.center.y), :], dpmm=self.image.dpmm
        )
        y_prof = SingleProfile(
            self.image[:, int(self.image.center.x)], dpmm=self.image.dpmm
        )
        x_width = x_prof.fwxm_data(x=fwxm)["width (exact) mm"]
        y_width = y_prof.fwxm_data(x=fwxm)["width (exact) mm"]
        if not np.allclose(x_width, y_width, atol=10):
            raise ValueError(
                f"The detected y and x field sizes were too different from one another. They should be within 1cm from each other. Detected field sizes: x={x_width}, y={y_width}"
            )
        if x_width > 140:
            return self.bb_positions_15x15
        else:
            return self.bb_positions_10x10

    def plot_analyzed_image(
        self, show: bool = True, **kwargs
    ) -> Tuple[List[plt.Figure], List[str]]:
        """Plot the analyzed image.

        Parameters
        ----------
        show : bool
            Whether to actually show the image when called.
        """
        figs = []
        names = []
        fig, axes = plt.subplots(1)
        figs.append(fig)
        names.append("image")
        self.image.plot(ax=axes, show=False, **kwargs)
        axes.axis("off")
        axes.set_title(f"{self.common_name} Phantom Analysis")

        # plot the bb marks
        for name, data in self.bb_centers.items():
            axes.plot(data.x, data.y, "go", label=name)

        # plot the bb center as small lines
        axes.axhline(
            y=self.bb_center.y, color="g", xmin=0.25, xmax=0.75, label="BB Centroid"
        )
        axes.axvline(x=self.bb_center.x, color="g", ymin=0.25, ymax=0.75)

        # plot the epid center as image-sized lines
        axes.axhline(y=self.epid_center.y, color="b", label="EPID Center")
        axes.axvline(x=self.epid_center.x, color="b")

        # plot the field center as field-sized lines
        axes.axhline(
            y=self.field_center.y,
            xmin=0.15,
            xmax=0.85,
            color="red",
            label="Field Center",
        )
        axes.axvline(x=self.field_center.x, ymin=0.15, ymax=0.85, color="red")

        axes.legend()

        if show:
            plt.show()
        return figs, names

    def save_analyzed_image(
        self,
        filename: Union[None, str, BinaryIO] = None,
        to_streams: bool = False,
        **kwargs,
    ) -> Optional[Union[Dict[str, BinaryIO], List[str]]]:
        """Save the analyzed image to disk or to stream. Kwargs are passed to plt.savefig()

        Parameters
        ----------
        filename : None, str, stream
            A string representing where to save the file to. If split_plots and to_streams are both true, leave as None as newly-created streams are returned.
        to_streams: bool
            This only matters if split_plots is True. If both of these are true, multiple streams will be created and returned as a dict.
        """
        if filename is None and to_streams is False:
            raise ValueError("Must pass in a filename unless saving to streams.")
        figs, names = self.plot_analyzed_image(show=False, **kwargs)
        if not to_streams:
            plt.savefig(filename, **kwargs)
        else:
            # append names to filename if it's file-like
            if not to_streams:
                filenames = []
                f, ext = osp.splitext(filename)
                for name in names:
                    filenames.append(f + "_" + name + ext)
            else:  # it's a stream buffer
                filenames = [io.BytesIO() for _ in names]
            for fig, name in zip(figs, filenames):
                fig.savefig(name, **kwargs)
            if to_streams:
                return {name: stream for name, stream in zip(names, filenames)}

    def publish_pdf(
        self,
        filename: str,
        notes: str = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ):
        """Publish (print) a PDF containing the analysis, images, and quantitative results.

        Parameters
        ----------
        filename : (str, file-like object}
            The file to write the results to.
        notes : str, list of strings
            Text; if str, prints single line.
            If list of strings, each list item is printed on its own line.
        open_file : bool
            Whether to open the file using the default program after creation.
        metadata : dict
            Extra data to be passed and shown in the PDF. The key and value will be shown with a colon.
            E.g. passing {'Author': 'James', 'Unit': 'TrueBeam'} would result in text in the PDF like:
            --------------
            Author: James
            Unit: TrueBeam
            --------------
        """
        canvas = pdf.PylinacCanvas(
            filename,
            page_title=f"{self.common_name} Phantom Analysis",
            metadata=metadata,
        )

        # write the text/numerical values
        text = self.results(as_list=True)
        canvas.add_text(text=text, location=(1.5, 25), font_size=14)
        if notes is not None:
            canvas.add_text(text="Notes:", location=(1, 5.5), font_size=12)
            canvas.add_text(text=notes, location=(1, 5))

        data = io.BytesIO()
        self.save_analyzed_image(data)
        canvas.add_image(data, location=(1, 3.5), dimensions=(19, 19))

        canvas.finish()
        if open_file:
            webbrowser.open(filename)


class IMTLRad(StandardImagingFC2):
    """The IMT light/rad phantom: https://www.imtqa.com/products/l-rad"""
    common_name = "IMT L-Rad"
    _demo_filename = "imtlrad.dcm"
    center_only_bb = {"Center": [0, 0]}
    bb_sampling_box_size_mm = 12
    field_strip_width_mm = 5

    def _determine_bb_set(self, fwxm: int) -> dict:
        return self.center_only_bb


class SNCFSQA(StandardImagingFC2):
    """SNC light/rad phantom. See the 'FSQA' phantom and specs: https://www.sunnuclear.com/products/suncheck-machine.

    Unlike other light/rad phantoms, this does not have at least a centered BB. The edge markers are in the penumbra
    and thus detecting them is difficult. We thus detect the one offset marker in the top right of the image.
    This is offset by 4cm in each direction. We can then assume that the phantom center is -4cm from this point,
    creating a 'virtual center' so we have an apples-to-apples comparison.
    """
    common_name = "SNC FSQA"
    _demo_filename = "FSQA_15x15.dcm"
    center_only_bb = {"TR": [40, -40]}
    bb_sampling_box_size_mm = 12
    field_strip_width_mm = 5

    def _determine_bb_set(self, fwxm: int) -> dict:
        return self.center_only_bb

    def _find_overall_bb_centroid(self, fwxm: int) -> Point:
        """Determine the geometric center of the 4 BBs"""
        # detect the upper right BB
        self.bb_centers = self._detect_bb_centers(fwxm)
        # add another virtual bb at the center of the phantom, knowing it's offset by 4cm in each direction
        self.bb_centers['Virtual Center'] = self.bb_centers['TR'] - Point(40*self.image.dpmm, -40*self.image.dpmm)
        return self.bb_centers['Virtual Center']


class LasVegas(ImagePhantomBase):
    _demo_filename = "lasvegas.dcm"
    common_name = "Las Vegas"
    phantom_bbox_size_mm2 = 20260
    detection_conditions = [is_centered, is_right_size]
    phantom_outline_object = {"Rectangle": {"width ratio": 0.62, "height ratio": 0.62}}
    low_contrast_background_roi_settings = {
        "roi 1": {"distance from center": 0.24, "angle": 0, "roi radius": 0.03},
        "roi 2": {"distance from center": 0.24, "angle": 90, "roi radius": 0.03},
        "roi 3": {"distance from center": 0.24, "angle": 180, "roi radius": 0.03},
        "roi 4": {"distance from center": 0.24, "angle": 270, "roi radius": 0.03},
    }
    low_contrast_roi_settings = {
        "roi 1": {"distance from center": 0.107, "angle": 0.5, "roi radius": 0.028},
        "roi 2": {"distance from center": 0.141, "angle": 39.5, "roi radius": 0.028},
        "roi 3": {"distance from center": 0.205, "angle": 58, "roi radius": 0.028},
        "roi 4": {"distance from center": 0.179, "angle": -76.5, "roi radius": 0.016},
        "roi 5": {"distance from center": 0.095, "angle": -63.5, "roi radius": 0.016},
        "roi 6": {"distance from center": 0.042, "angle": 0.5, "roi radius": 0.016},
        "roi 7": {"distance from center": 0.097, "angle": 65.5, "roi radius": 0.016},
        "roi 8": {"distance from center": 0.178, "angle": 76.5, "roi radius": 0.016},
        "roi 9": {"distance from center": 0.174, "angle": -97.5, "roi radius": 0.012},
        "roi 10": {"distance from center": 0.088, "angle": -105.5, "roi radius": 0.012},
        "roi 11": {"distance from center": 0.024, "angle": -183.5, "roi radius": 0.012},
        "roi 12": {"distance from center": 0.091, "angle": 105.5, "roi radius": 0.012},
        "roi 13": {"distance from center": 0.179, "angle": 97.5, "roi radius": 0.012},
        "roi 14": {"distance from center": 0.189, "angle": -113.5, "roi radius": 0.007},
        "roi 15": {"distance from center": 0.113, "angle": -131.5, "roi radius": 0.007},
        "roi 16": {
            "distance from center": 0.0745,
            "angle": -181.5,
            "roi radius": 0.007,
        },
        "roi 17": {"distance from center": 0.115, "angle": 130, "roi radius": 0.007},
        "roi 18": {"distance from center": 0.191, "angle": 113, "roi radius": 0.007},
        "roi 19": {
            "distance from center": 0.2085,
            "angle": -124.6,
            "roi radius": 0.003,
        },
        "roi 20": {"distance from center": 0.146, "angle": -144.3, "roi radius": 0.003},
    }

    @staticmethod
    def run_demo():
        """Run the Las Vegas phantom analysis demonstration."""
        lv = LasVegas.from_demo_image()
        lv.analyze()
        lv.plot_analyzed_image()

    def _preprocess(self):
        self._check_direction()

    def _check_inversion(self):
        """Check the inversion by using the histogram of the phantom region"""
        roi = self.phantom_ski_region
        phantom_array = self.image.array[
            roi.bbox[0] : roi.bbox[2], roi.bbox[1] : roi.bbox[3]
        ]
        phantom_sub_image = image.load(phantom_array)
        phantom_sub_image.crop(int(phantom_sub_image.shape[0] * 0.1))
        p5 = np.percentile(phantom_sub_image, 0.5)
        p50 = np.percentile(phantom_sub_image, 50)
        p95 = np.percentile(phantom_sub_image, 99.5)
        dist_to_5 = abs(p50 - p5)
        dist_to_95 = abs(p50 - p95)
        if dist_to_5 > dist_to_95:
            self.image.invert()

    def _check_direction(self) -> None:
        """Check that the phantom is facing the right direction and if not perform a left-right flip of the array."""
        circle = CollapsedCircleProfile(
            self.phantom_center,
            self.phantom_radius * 0.175,
            self.image,
            ccw=False,
            width_ratio=0.16,
            num_profiles=5,
        )
        roll_amount = np.where(circle.values == circle.values.min())[0][0]
        circle.roll(roll_amount)
        circle.filter(size=0.015, kind="median")
        valley_idxs, _ = circle.find_peaks(max_number=2)
        if valley_idxs[0] > valley_idxs[1]:
            self.image.array = np.fliplr(self.image.array)
            self._phantom_ski_region = None

    def _phantom_radius_calc(self) -> float:
        return math.sqrt(self.phantom_ski_region.area_bbox) * 1.626

    def _phantom_angle_calc(self) -> float:
        return 0.0

    def _plot_lowcontrast_graph(self, axes: plt.Axes):
        """Plot the low contrast ROIs to an axes, including visibility"""
        # plot contrast
        (line1,) = axes.plot(
            [roi.contrast for roi in self.low_contrast_rois],
            marker="o",
            color="m",
            label="Contrast",
        )
        axes.axhline(self._low_contrast_threshold, color="m")
        axes.grid(True)
        axes.set_title("Low-frequency Contrast")
        axes.set_xlabel("ROI #")
        axes.set_ylabel("Contrast")
        # plot CNR
        axes2 = axes.twinx()
        axes2.set_ylabel("CNR")
        (line2,) = axes2.plot(
            [roi.contrast_to_noise for roi in self.low_contrast_rois],
            marker="^",
            label="CNR",
        )
        # plot visibility; here's what different from the base method
        axes3 = axes.twinx()
        axes3.set_ylabel("Visibility")
        (line3,) = axes3.plot(
            [roi.visibility for roi in self.low_contrast_rois],
            marker="*",
            color="blue",
            label="Visibility",
        )
        axes3.axhline(self.visibility_threshold, color="blue")
        axes3.spines.right.set_position(("axes", 1.2))

        axes.legend(handles=[line1, line2, line3])

    def results(self) -> str:
        """Return the results of the analysis. Overridden because ROIs seen is based on visibility, not CNR"""
        text = [f"{self.common_name} results:", f"File: {self.image.truncated_path}"]
        text += [
            f"Median Contrast: {np.median([roi.contrast for roi in self.low_contrast_rois]):2.2f}",
            f"Median CNR: {np.median([roi.contrast_to_noise for roi in self.low_contrast_rois]):2.1f}",
            f'# Low contrast ROIs "seen": {sum(roi.passed_visibility for roi in self.low_contrast_rois):2.0f} of {len(self.low_contrast_rois)}',
        ]
        text = "\n".join(text)
        return text

    def results_data(self, as_dict=False) -> Union[PlanarResult, dict]:
        """Overridden because ROIs seen is based on visibility, not CNR"""
        data = PlanarResult(
            analysis_type=self.common_name,
            median_contrast=np.median([roi.contrast for roi in self.low_contrast_rois]),
            median_cnr=np.median(
                [roi.contrast_to_noise for roi in self.low_contrast_rois]
            ),
            num_contrast_rois_seen=sum(
                roi.passed_visibility for roi in self.low_contrast_rois
            ),
            phantom_center_x_y=(self.phantom_center.x, self.phantom_center.y),
        )
        if as_dict:
            return dataclasses.asdict(data)
        return data


class PTWEPIDQC(ImagePhantomBase):
    _demo_filename = "PTW-EPID-QC.dcm"
    common_name = "PTW EPID QC"
    phantom_bbox_size_mm2 = 250**2
    detection_conditions = [is_centered, is_right_size]
    detection_canny_settings = {"sigma": 4, "percentiles": (0.001, 0.01)}
    phantom_outline_object = {"Rectangle": {"width ratio": 8.55, "height ratio": 8.55}}
    high_contrast_roi_settings = {
        # angled rois
        "roi 1": {
            "distance from center": 1.5,
            "angle": -135,
            "roi radius": 0.35,
            "lp/mm": 0.15,
        },
        "roi 2": {
            "distance from center": 3.1,
            "angle": -109,
            "roi radius": 0.35,
            "lp/mm": 0.21,
        },
        "roi 3": {
            "distance from center": 3.4,
            "angle": -60,
            "roi radius": 0.3,
            "lp/mm": 0.27,
        },
        "roi 4": {
            "distance from center": 1.9,
            "angle": -60,
            "roi radius": 0.25,
            "lp/mm": 0.33,
        },
        # vertical rois
        "roi 5": {
            "distance from center": 3.68,
            "angle": -90,
            "roi radius": 0.18,
            "lp/mm": 0.5,
        },
        "roi 6": {
            "distance from center": 2.9,
            "angle": -90,
            "roi radius": 0.08,
            "lp/mm": 2,
        },
        "roi 7": {
            "distance from center": 2.2,
            "angle": -90,
            "roi radius": 0.04,
            "lp/mm": 3,
        },
    }
    low_contrast_roi_settings = {
        "roi 1": {"distance from center": 3.87, "angle": 31, "roi radius": 0.3},
        "roi 2": {"distance from center": 3.48, "angle": 17, "roi radius": 0.3},
        "roi 3": {"distance from center": 3.3, "angle": 0, "roi radius": 0.3},
        "roi 4": {"distance from center": 3.48, "angle": -17, "roi radius": 0.3},
        "roi 5": {"distance from center": 3.87, "angle": -31, "roi radius": 0.3},
        "roi 6": {"distance from center": 3.87, "angle": 180 - 31, "roi radius": 0.3},
        "roi 7": {"distance from center": 3.48, "angle": 180 - 17, "roi radius": 0.3},
        "roi 8": {"distance from center": 3.3, "angle": 180, "roi radius": 0.3},
        "roi 9": {"distance from center": 3.48, "angle": 180 + 17, "roi radius": 0.3},
    }
    low_contrast_background_roi_settings = {
        "roi 1": {"distance from center": 3.85, "angle": -148, "roi radius": 0.3},
    }

    @staticmethod
    def run_demo() -> None:
        """Run the Standard Imaging QC-3 phantom analysis demonstration."""
        ptw = PTWEPIDQC.from_demo_image()
        ptw.analyze()
        ptw.plot_analyzed_image()

    def _phantom_radius_calc(self) -> float:
        """The radius of the phantom in pixels; the value itself doesn't matter, it's just
        used for relative distances to ROIs.

        Returns
        -------
        radius : float
        """
        return math.sqrt(self.phantom_ski_region.area_bbox) * 0.116

    def _phantom_angle_calc(self) -> float:
        """The angle of the phantom. This assumes the user has placed the phantom with the high-contrast line pairs at the top
        and low contrast at the bottom at a fixed rotation of 0 degrees.

        Returns
        -------
        angle : float
            The angle in degrees.
        """
        return 0


class StandardImagingQC3(ImagePhantomBase):
    _demo_filename = "qc3.dcm"
    common_name = "SI QC-3"
    phantom_bbox_size_mm2 = 168**2
    detection_conditions = [is_centered, is_right_size]
    phantom_outline_object = {"Rectangle": {"width ratio": 7.5, "height ratio": 6}}
    high_contrast_roi_settings = {
        "roi 1": {
            "distance from center": 2.8,
            "angle": 0,
            "roi radius": 0.5,
            "lp/mm": 0.1,
        },
        "roi 2": {
            "distance from center": -2.8,
            "angle": 0,
            "roi radius": 0.5,
            "lp/mm": 0.2,
        },
        "roi 3": {
            "distance from center": 1.45,
            "angle": 0,
            "roi radius": 0.5,
            "lp/mm": 0.25,
        },
        "roi 4": {
            "distance from center": -1.45,
            "angle": 0,
            "roi radius": 0.5,
            "lp/mm": 0.45,
        },
        "roi 5": {
            "distance from center": 0,
            "angle": 0,
            "roi radius": 0.5,
            "lp/mm": 0.76,
        },
    }
    low_contrast_roi_settings = {
        "roi 1": {"distance from center": 2, "angle": -90, "roi radius": 0.5},
        "roi 2": {"distance from center": 2.4, "angle": 55, "roi radius": 0.5},
        "roi 3": {"distance from center": 2.4, "angle": -55, "roi radius": 0.5},
        "roi 4": {"distance from center": 2.4, "angle": 128, "roi radius": 0.5},
        "roi 5": {"distance from center": 2.4, "angle": -128, "roi radius": 0.5},
    }
    low_contrast_background_roi_settings = {
        "roi 1": {"distance from center": 2, "angle": 90, "roi radius": 0.5},
    }

    @classmethod
    def from_demo_image(cls):
        """Instantiate and load the demo image."""
        demo_file = retrieve_demo_file(name=cls._demo_filename)
        inst = cls(demo_file)
        inst.image.invert()
        return inst

    @staticmethod
    def run_demo() -> None:
        """Run the Standard Imaging QC-3 phantom analysis demonstration."""
        qc3 = StandardImagingQC3.from_demo_image()
        qc3.analyze()
        qc3.plot_analyzed_image()

    def _phantom_radius_calc(self) -> float:
        """The radius of the phantom in pixels; the value itself doesn't matter, it's just
        used for relative distances to ROIs.

        Returns
        -------
        radius : float
        """
        return math.sqrt(self.phantom_ski_region.area_bbox) * 0.0896

    @lru_cache()
    def _phantom_angle_calc(self) -> float:
        """The angle of the phantom. This assumes the user is using the stand that comes with the phantom,
        which angles the phantom at 45 degrees.

        Returns
        -------
        angle : float
            The angle in degrees.
        """
        angle = np.degrees(self.phantom_ski_region.orientation)
        if np.isclose(angle, 45, atol=5):
            return 45
        elif np.isclose(angle, -45, atol=5):
            return -45
        else:
            raise ValueError(
                "The phantom angle was not near +/-45 degrees. Please adjust the phantom."
            )


class StandardImagingQCkV(StandardImagingQC3):
    _demo_filename = "SI-QC-kV.dcm"
    common_name = "SI QC-kV"
    phantom_bbox_size_mm2 = 142**2
    detection_conditions = [is_centered, is_right_size]
    phantom_outline_object = {"Rectangle": {"width ratio": 7.8, "height ratio": 6.4}}
    high_contrast_roi_settings = {
        "roi 1": {
            "distance from center": 2.8,
            "angle": 0,
            "roi radius": 0.5,
            "lp/mm": 0.66,
        },
        "roi 2": {
            "distance from center": -2.8,
            "angle": 0,
            "roi radius": 0.5,
            "lp/mm": 0.98,
        },
        "roi 3": {
            "distance from center": 1.45,
            "angle": 0,
            "roi radius": 0.5,
            "lp/mm": 1.50,
        },
        "roi 4": {
            "distance from center": -1.45,
            "angle": 0,
            "roi radius": 0.5,
            "lp/mm": 2.00,
        },
        "roi 5": {
            "distance from center": 0,
            "angle": 0,
            "roi radius": 0.5,
            "lp/mm": 2.46,
        },
    }
    low_contrast_roi_settings = {
        "roi 1": {"distance from center": 2, "angle": -90, "roi radius": 0.5},
        "roi 2": {"distance from center": 2.4, "angle": 55, "roi radius": 0.5},
        "roi 3": {"distance from center": 2.4, "angle": -55, "roi radius": 0.5},
        "roi 4": {"distance from center": 2.4, "angle": 128, "roi radius": 0.5},
        "roi 5": {"distance from center": 2.4, "angle": -128, "roi radius": 0.5},
    }
    low_contrast_background_roi_settings = {
        "roi 1": {"distance from center": 2, "angle": 90, "roi radius": 0.5},
    }

    @staticmethod
    def run_demo() -> None:
        """Run the Standard Imaging QC-3 phantom analysis demonstration."""
        qc3 = StandardImagingQCkV.from_demo_image()
        qc3.analyze()
        qc3.plot_analyzed_image()

    def _phantom_radius_calc(self) -> float:
        """The radius of the phantom in pixels; the value itself doesn't matter, it's just
        used for relative distances to ROIs.

        Returns
        -------
        radius : float
        """
        return math.sqrt(self.phantom_ski_region.area_bbox) * 0.0989


class SNCkV(ImagePhantomBase):
    _demo_filename = "SNC-kV.dcm"
    common_name = "SNC kV-QA"
    phantom_bbox_size_mm2 = 134**2
    detection_conditions = [is_centered, is_right_size]
    phantom_outline_object = {"Rectangle": {"width ratio": 7.7, "height ratio": 5.6}}
    high_contrast_roi_settings = {
        "roi 1": {
            "distance from center": 1.8,
            "angle": 0,
            "roi radius": 0.7,
            "lp/mm": 0.6,
        },
        "roi 2": {
            "distance from center": -1.8,
            "angle": 90,
            "roi radius": 0.7,
            "lp/mm": 1.2,
        },
        "roi 3": {
            "distance from center": -1.8,
            "angle": 0,
            "roi radius": 0.7,
            "lp/mm": 1.8,
        },
        "roi 4": {
            "distance from center": 1.8,
            "angle": 90,
            "roi radius": 0.7,
            "lp/mm": 2.4,
        },
    }
    low_contrast_roi_settings = {
        "roi 1": {"distance from center": 2.6, "angle": -45, "roi radius": 0.6},
        "roi 2": {"distance from center": 2.6, "angle": -135, "roi radius": 0.6},
        "roi 3": {"distance from center": 2.6, "angle": 45, "roi radius": 0.6},
        "roi 4": {"distance from center": 2.6, "angle": 135, "roi radius": 0.6},
    }
    low_contrast_background_roi_settings = {
        "roi 1": {"distance from center": 0.5, "angle": 90, "roi radius": 0.25},
        "roi 2": {"distance from center": 0.5, "angle": -90, "roi radius": 0.25},
    }

    @staticmethod
    def run_demo() -> None:
        """Run the Sun Nuclear kV-QA phantom analysis demonstration."""
        snc = SNCkV.from_demo_image()
        snc.analyze()
        snc.plot_analyzed_image()

    def _phantom_radius_calc(self) -> float:
        """The radius of the phantom in pixels; the value itself doesn't matter, it's just
        used for relative distances to ROIs.

        Returns
        -------
        radius : float
        """
        return math.sqrt(self.phantom_ski_region.area_bbox) * 0.1071

    def _phantom_angle_calc(self) -> float:
        """The angle of the phantom. This assumes the user is using the stand that comes with the phantom,
        which angles the phantom at 45 degrees.

        Returns
        -------
        angle : float
            The angle in degrees.
        """
        return 135


class SNCMV(SNCkV):
    _demo_filename = "SNC-MV.dcm"
    common_name = "SNC MV-QA"
    phantom_bbox_size_mm2 = 118**2
    detection_conditions = [is_centered, is_right_size]
    phantom_outline_object = {"Rectangle": {"width ratio": 7.5, "height ratio": 7.5}}
    high_contrast_roi_settings = {
        "roi 1": {
            "distance from center": -2.3,
            "angle": 0,
            "roi radius": 0.8,
            "lp/mm": 0.1,
        },
        "roi 2": {
            "distance from center": 2.3,
            "angle": 90,
            "roi radius": 0.8,
            "lp/mm": 0.2,
        },
        "roi 3": {
            "distance from center": 2.3,
            "angle": 0,
            "roi radius": 0.8,
            "lp/mm": 0.5,
        },
        "roi 4": {
            "distance from center": -2.3,
            "angle": 90,
            "roi radius": 0.8,
            "lp/mm": 1.0,
        },
    }
    low_contrast_roi_settings = {
        "roi 1": {"distance from center": 3.4, "angle": -45, "roi radius": 0.7},
        "roi 2": {"distance from center": 3.4, "angle": 45, "roi radius": 0.7},
        "roi 3": {"distance from center": 3.4, "angle": 135, "roi radius": 0.7},
        "roi 4": {"distance from center": 3.4, "angle": -135, "roi radius": 0.7},
    }
    low_contrast_background_roi_settings = {
        "roi 1": {"distance from center": 0.7, "angle": 0, "roi radius": 0.2},
        "roi 2": {"distance from center": -0.7, "angle": 0, "roi radius": 0.2},
    }

    @staticmethod
    def run_demo() -> None:
        """Run the Sun Nuclear MV-QA phantom analysis demonstration."""
        snc = SNCMV.from_demo_image()
        snc.analyze()
        snc.plot_analyzed_image()

    def _phantom_angle_calc(self) -> float:
        """The angle of the phantom. This assumes the user is using the stand that comes with the phantom,
        which angles the phantom at 45 degrees.

        Returns
        -------
        angle : float
            The angle in degrees.
        """
        return 45


class LeedsTOR(ImagePhantomBase):
    _demo_filename = "leeds.dcm"
    common_name = "Leeds"
    phantom_bbox_size_mm2 = 148**2
    _is_ccw = False
    phantom_outline_object = {"Circle": {"radius ratio": 0.97}}
    high_contrast_roi_settings = {
        "roi 1": {
            "distance from center": 0.3,
            "angle": 54.8,
            "roi radius": 0.04,
            "lp/mm": 0.5,
        },
        "roi 2": {
            "distance from center": 0.187,
            "angle": 25.1,
            "roi radius": 0.04,
            "lp/mm": 0.56,
        },
        "roi 3": {
            "distance from center": 0.187,
            "angle": -27.5,
            "roi radius": 0.04,
            "lp/mm": 0.63,
        },
        "roi 4": {
            "distance from center": 0.252,
            "angle": 79.7,
            "roi radius": 0.03,
            "lp/mm": 0.71,
        },
        "roi 5": {
            "distance from center": 0.092,
            "angle": 63.4,
            "roi radius": 0.03,
            "lp/mm": 0.8,
        },
        "roi 6": {
            "distance from center": 0.094,
            "angle": -65,
            "roi radius": 0.02,
            "lp/mm": 0.9,
        },
        "roi 7": {
            "distance from center": 0.252,
            "angle": -263,
            "roi radius": 0.02,
            "lp/mm": 1.0,
        },
        "roi 8": {
            "distance from center": 0.094,
            "angle": -246,
            "roi radius": 0.018,
            "lp/mm": 1.12,
        },
        "roi 9": {
            "distance from center": 0.0958,
            "angle": -117,
            "roi radius": 0.018,
            "lp/mm": 1.25,
        },
        "roi 10": {
            "distance from center": 0.27,
            "angle": 112.5,
            "roi radius": 0.015,
            "lp/mm": 1.4,
        },
        "roi 11": {
            "distance from center": 0.13,
            "angle": 145,
            "roi radius": 0.015,
            "lp/mm": 1.6,
        },
        "roi 12": {
            "distance from center": 0.135,
            "angle": -142,
            "roi radius": 0.011,
            "lp/mm": 1.8,
        },
    }
    low_contrast_background_roi_settings = {
        "roi 1": {"distance from center": 0.65, "angle": 30, "roi radius": 0.025},
        "roi 2": {"distance from center": 0.65, "angle": 120, "roi radius": 0.025},
        "roi 3": {"distance from center": 0.65, "angle": 210, "roi radius": 0.025},
        "roi 4": {"distance from center": 0.65, "angle": 300, "roi radius": 0.025},
    }
    low_contrast_roi_settings = {
        # set 1
        "roi 1": {"distance from center": 0.785, "angle": 30, "roi radius": 0.025},
        "roi 2": {"distance from center": 0.785, "angle": 45, "roi radius": 0.025},
        "roi 3": {"distance from center": 0.785, "angle": 60, "roi radius": 0.025},
        "roi 4": {"distance from center": 0.785, "angle": 75, "roi radius": 0.025},
        "roi 5": {"distance from center": 0.785, "angle": 90, "roi radius": 0.025},
        "roi 6": {"distance from center": 0.785, "angle": 105, "roi radius": 0.025},
        "roi 7": {"distance from center": 0.785, "angle": 120, "roi radius": 0.025},
        "roi 8": {"distance from center": 0.785, "angle": 135, "roi radius": 0.025},
        "roi 9": {"distance from center": 0.785, "angle": 150, "roi radius": 0.025},
        # set 2
        "roi 10": {"distance from center": 0.785, "angle": 210, "roi radius": 0.025},
        "roi 11": {"distance from center": 0.785, "angle": 225, "roi radius": 0.025},
        "roi 12": {"distance from center": 0.785, "angle": 240, "roi radius": 0.025},
        "roi 13": {"distance from center": 0.785, "angle": 255, "roi radius": 0.025},
        "roi 14": {"distance from center": 0.785, "angle": 270, "roi radius": 0.025},
        "roi 15": {"distance from center": 0.785, "angle": 285, "roi radius": 0.025},
        "roi 16": {"distance from center": 0.785, "angle": 300, "roi radius": 0.025},
        "roi 17": {"distance from center": 0.785, "angle": 315, "roi radius": 0.025},
        "roi 18": {"distance from center": 0.785, "angle": 330, "roi radius": 0.025},
    }

    @staticmethod
    def run_demo() -> None:
        """Run the Leeds TOR phantom analysis demonstration."""
        leeds = LeedsTOR.from_demo_image()
        leeds.analyze()
        leeds.plot_analyzed_image()

    @lru_cache()
    def _phantom_angle_calc(self) -> float:
        """Determine the angle of the phantom.

        This is done by searching for square-like boxes of the canny image. There are usually two: one lead and
        one copper. The box with the highest intensity (lead) is identified. The angle from the center of the lead
        square bounding box and the phantom center determines the phantom angle.

        Returns
        -------
        angle : float
            The angle in degrees
        """
        start_angle_deg = self._determine_start_angle_for_circle_profile()
        circle = self._circle_profile_for_phantom_angle(start_angle_deg, is_ccw=True)
        peak_idx, _ = circle.find_fwxm_peaks(threshold=0.6, max_number=1)

        shift_percent = peak_idx[0] / len(circle.values)
        shift_radians = shift_percent * 2 * np.pi
        shift_radians_corrected = 2 * np.pi - shift_radians

        angle = np.degrees(shift_radians_corrected) + start_angle_deg
        return angle

    def _phantom_radius_calc(self) -> float:
        """Determine the radius of the phantom.

        The radius is determined by finding the largest of the detected blobs of the canny image and taking
        its major axis length.

        Returns
        -------
        radius : float
            The radius of the phantom in pixels. The actual value is not important; it is used for scaling the
            distances to the low and high contrast ROIs.
        """
        return math.sqrt(self.phantom_ski_region.area_bbox) * 0.515

    def _determine_start_angle_for_circle_profile(self) -> float:
        """Determine an appropriate angle for starting the circular profile
        used to determine the phantom angle.

        In most cases we can just use 0 degs but for the case where the phantom
        is set up near 0 degs, the peak of the circular profile will be split
        between the left and right sides of the profile.  We can check for this
        case by looking at a few of the peak indexes and determining whether
        they are all on the left or right side of the profile or split left and
        right.  If they're split left and right, then we we need to use a
        different circular profile start angle  to get an accurate angle
        determination

        Returns
        -------
        start_angle_deg: float
            The start angle to be used for the circular profile used to determine the
            phantom rotation.
        """

        circle = self._circle_profile_for_phantom_angle(0)
        peak_idxs, _ = circle.find_fwxm_peaks(threshold=0.6, max_number=4)
        on_left_half = [x < len(circle.values) / 2 for x in peak_idxs]
        aligned_to_zero_deg = not (all(on_left_half) or not any(on_left_half))
        return 90 if aligned_to_zero_deg else 0

    def _preprocess(self) -> None:
        self._check_if_counter_clockwise()

    def _check_if_counter_clockwise(self) -> None:
        """Determine if the low-contrast bubbles go from high to low clockwise or counter-clockwise."""
        circle = self._circle_profile_for_phantom_angle(0)
        peak_idx, _ = circle.find_fwxm_peaks(threshold=0.6, max_number=1)
        circle.values = np.roll(circle.values, -peak_idx[0])
        _, first_set = circle.find_peaks(
            search_region=(0.05, 0.45), threshold=0, min_distance=0.025, max_number=9
        )
        _, second_set = circle.find_peaks(
            search_region=(0.55, 0.95), threshold=0, min_distance=0.025, max_number=9
        )
        self._is_ccw = max(first_set) > max(second_set)
        if not self._is_ccw:
            self.image.fliplr()
            del (
                self.phantom_ski_region
            )  # clear the property to calculate it again since we flipped it

    def _circle_profile_for_phantom_angle(
        self, start_angle_deg: float, is_ccw: bool = False
    ) -> CollapsedCircleProfile:
        """Create a circular profile centered at phantom origin

        Parameters
        ----------
        start_angle_deg: float

            Angle in degrees at which to start the profile
        Returns
        -------
        circle : CollapsedCircleProfile
            The circular profile centered on the phantom center and origin set to the given start angle.
        """
        circle = CollapsedCircleProfile(
            self.phantom_center,
            self.phantom_radius * 0.79,
            self.image.array,
            width_ratio=0.04,
            ccw=is_ccw,
            start_angle=np.deg2rad(start_angle_deg),
        )
        circle.ground()
        circle.filter(size=0.01)
        circle.invert()
        return circle


class DoselabMC2kV(ImagePhantomBase):
    common_name = "Doselab MC2 kV"
    _demo_filename = "Doselab_kV.dcm"
    phantom_bbox_size_mm2 = 26300
    detection_conditions = [is_right_size]
    phantom_outline_object = {"Rectangle": {"width ratio": 0.55, "height ratio": 0.63}}
    low_contrast_background_roi_settings = {
        "roi 1": {"distance from center": 0.27, "angle": 48.5, "roi radius": 0.025},
    }
    low_contrast_roi_settings = {
        "roi 1": {"distance from center": 0.27, "angle": -48.5, "roi radius": 0.025},
        "roi 2": {"distance from center": 0.225, "angle": -65, "roi radius": 0.025},
        "roi 3": {"distance from center": 0.205, "angle": -88.5, "roi radius": 0.025},
        "roi 4": {"distance from center": 0.22, "angle": -110, "roi radius": 0.025},
        "roi 5": {"distance from center": 0.22, "angle": 110, "roi radius": 0.025},
        "roi 6": {"distance from center": 0.205, "angle": 88.5, "roi radius": 0.025},
        "roi 7": {"distance from center": 0.225, "angle": 65, "roi radius": 0.025},
    }
    high_contrast_roi_settings = {
        "roi 1": {
            "distance from center": 0.17,
            "angle": -20,
            "roi radius": 0.013,
            "lp/mm": 0.6,
        },
        "roi 2": {
            "distance from center": 0.16,
            "angle": -2,
            "roi radius": 0.007,
            "lp/mm": 1.2,
        },
        "roi 3": {
            "distance from center": 0.164,
            "angle": 12.8,
            "roi radius": 0.005,
            "lp/mm": 1.8,
        },
        "roi 4": {
            "distance from center": 0.175,
            "angle": 24.7,
            "roi radius": 0.0035,
            "lp/mm": 2.4,
        },
    }

    @staticmethod
    def run_demo() -> None:
        """Run the Doselab MC2 kV-area phantom analysis demonstration."""
        leeds = DoselabMC2kV.from_demo_image()
        leeds.analyze()
        leeds.plot_analyzed_image()

    def _phantom_radius_calc(self) -> float:
        return math.sqrt(self.phantom_ski_region.area_bbox) * 1.214

    def _phantom_angle_calc(self) -> float:
        roi = self.phantom_ski_region
        angle = np.degrees(roi.orientation) + 90
        if not np.isclose(angle, 45, atol=5):
            raise ValueError(
                "Angles not close enough to the ideal 45 degrees. Check phantom setup or override angle."
            )
        return angle


class DoselabMC2MV(DoselabMC2kV):
    common_name = "Doselab MC2 MV"
    _demo_filename = "Doselab_MV.dcm"
    low_contrast_background_roi_settings = {
        "roi 1": {"distance from center": 0.27, "angle": 48.5, "roi radius": 0.025},
    }
    low_contrast_roi_settings = {
        "roi 1": {"distance from center": 0.27, "angle": -48.5, "roi radius": 0.025},
        "roi 2": {"distance from center": 0.225, "angle": -65, "roi radius": 0.025},
        "roi 3": {"distance from center": 0.205, "angle": -88.5, "roi radius": 0.025},
        "roi 4": {"distance from center": 0.22, "angle": -110, "roi radius": 0.025},
        "roi 5": {"distance from center": 0.22, "angle": 110, "roi radius": 0.025},
        "roi 6": {"distance from center": 0.205, "angle": 88.5, "roi radius": 0.025},
        "roi 7": {"distance from center": 0.225, "angle": 65, "roi radius": 0.025},
    }
    high_contrast_roi_settings = {
        "roi 1": {
            "distance from center": 0.23,
            "angle": -135.3,
            "roi radius": 0.012,
            "lp/mm": 0.1,
        },
        "roi 2": {
            "distance from center": 0.173,
            "angle": 161,
            "roi radius": 0.012,
            "lp/mm": 0.2,
        },
        "roi 3": {
            "distance from center": 0.237,
            "angle": 133,
            "roi radius": 0.012,
            "lp/mm": 0.4,
        },
        "roi 4": {
            "distance from center": 0.298,
            "angle": 122.9,
            "roi radius": 0.01,
            "lp/mm": 0.8,
        },
    }

    @staticmethod
    def run_demo() -> None:
        """Run the Doselab MC2 MV-area phantom analysis demonstration."""
        leeds = DoselabMC2MV.from_demo_image()
        leeds.analyze()
        leeds.plot_analyzed_image()


def take_centermost_roi(rprops: List[RegionProperties], image_shape: Tuple[int, int]):
    """Return the ROI that is closest to the center."""
    larger_rois = [
        rprop for rprop in rprops if rprop.area > 20 and rprop.eccentricity < 0.9
    ]  # drop stray pixel ROIs and line-like ROIs
    center_roi = sorted(
        larger_rois,
        key=lambda p: abs(p.centroid[0] - image_shape[0] / 2)
        + abs(p.centroid[1] - image_shape[1] / 2),
    )[0]
    return center_roi



import functools
import weakref


def lru_cache(*lru_args, **lru_kwargs):
    """Method-safe LRU cache; https://stackoverflow.com/a/33672499"""

    def decorator(func):
        @functools.wraps(func)
        def wrapped_func(self, *args, **kwargs):
            # We're storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            self_weak = weakref.ref(self)

            @functools.wraps(func)
            @functools.lru_cache(*lru_args, **lru_kwargs)
            def cached_method(*args, **kwargs):
                return func(self_weak(), *args, **kwargs)

            setattr(self, func.__name__, cached_method)
            return cached_method(*args, **kwargs)

        return wrapped_func

    return decorator



from enum import Enum

import argue


def noop(value: float) -> float:
    """Don't do anything."""
    return value


def mirror_360(value: float) -> float:
    """Mirror about 0"""
    argue.verify_bounds(value, argue.POSITIVE)
    return 360 - value


def shift_and_mirror_360(value: float) -> float:
    """Shift by 180 degrees and then mirror about 0"""
    argue.verify_bounds(value, argue.POSITIVE)
    v = value - 180
    if v > 0:
        return mirror_360(v)
    else:
        return abs(v)


def inv_shift_and_mirror_360(value: float) -> float:
    """Inverse shift and mirror"""
    v = 180 - value
    if v < 0:
        return mirror_360(abs(v))
    else:
        return v


class MachineScale(Enum):
    """Possible machine scales. Used for specifying input and output scales for conversion.
    The enum keys are conversion functions for each axis relative to IEC 61217"""

    IEC61217 = {
        "gantry_to_iec": noop,
        "collimator_to_iec": noop,
        "rotation_to_iec": noop,
        "gantry_from_iec": noop,
        "collimator_from_iec": noop,
        "rotation_from_iec": noop,
    }
    ELEKTA_IEC = {
        "gantry_to_iec": noop,
        "collimator_to_iec": noop,
        "rotation_to_iec": mirror_360,
        "gantry_from_iec": noop,
        "collimator_from_iec": noop,
        "rotation_from_iec": mirror_360,
    }
    VARIAN_IEC = {
        "gantry_to_iec": noop,
        "collimator_to_iec": noop,
        "rotation_to_iec": mirror_360,
        "gantry_from_iec": noop,
        "collimator_from_iec": noop,
        "rotation_from_iec": mirror_360,
    }
    VARIAN_STANDARD = {
        "gantry_to_iec": shift_and_mirror_360,
        "collimator_to_iec": shift_and_mirror_360,
        "rotation_to_iec": shift_and_mirror_360,
        "gantry_from_iec": inv_shift_and_mirror_360,
        "collimator_from_iec": inv_shift_and_mirror_360,
        "rotation_from_iec": inv_shift_and_mirror_360,
    }


def convert(
    input_scale: MachineScale,
    output_scale: MachineScale,
    gantry: float,
    collimator: float,
    rotation: float,
) -> (float, float, float):
    """Convert from one coordinate scale to another. Returns gantry, collimator, rotation."""
    # convert to IEC61217 since everything is defined relative to it
    g = input_scale.value["gantry_to_iec"](gantry)
    c = input_scale.value["collimator_to_iec"](collimator)
    r = input_scale.value["rotation_to_iec"](rotation)
    # now apply the inverse to go from 61217 to output scale
    g_out = output_scale.value["gantry_from_iec"](g)
    c_out = output_scale.value["collimator_from_iec"](c)
    r_out = output_scale.value["rotation_from_iec"](r)
    return g_out, c_out, r_out



"""Module for processing "masked" arrays, i.e. binary images."""

import numpy as np


def bounding_box(array: np.array) -> (float, ...):
    # TODO: replace with regionprops
    """Get the bounding box values of an ROI in a 2D array."""
    binary_arr = np.argwhere(array)
    (ymin, xmin), (ymax, xmax) = binary_arr.min(0), binary_arr.max(0) + 1
    return ymin, ymax, xmin, xmax



"""Pylinac settings"""
from matplotlib.pyplot import cm

# use a string or colormap option. See options here: http://matplotlib.org/examples/color/colormaps_reference.html
DICOM_COLORMAP = cm.gray
ARRAY_COLORMAP = cm.gray
PATH_TRUNCATION_LENGTH = 80


def get_dicom_cmap():
    """Return the DICOM colormap. Passed to cmap parameter in matplotlib calls."""
    return DICOM_COLORMAP


def get_array_cmap():
    """Return the array colormap. Passed to cmap parameter in matplotlib calls."""
    return ARRAY_COLORMAP



"""Perform non-linear regression using a Hill function."""
from __future__ import annotations

import math

import numpy as np
from scipy.optimize import curve_fit


class Hill:
    """A Hill function is for modeling a field falloff (i.e. penumbra). It's not a perfect model, but it fits to a
    function, which is not limited by resolution issues as may be experienced on low-res devices like ion chamber arrays."""

    @classmethod
    def fit(cls, x_data: np.ndarray, y_data: np.ndarray) -> Hill:
        """Fit x & y data to a Hill function."""
        fitted_parameters, _ = curve_fit(
            hill_func,
            x_data,
            y_data,
            p0=(min(y_data), max(y_data), np.median(x_data), 0),
        )
        instance = cls()
        instance.params = fitted_parameters
        return instance

    def inflection_idx(self) -> dict:
        """Determine the x-value inflection point of the fitted Hill function."""
        idx = self.params[2] * math.pow(
            (self.params[3] - 1) / (self.params[3] + 1), 1 / self.params[3]
        )
        return {"index (exact)": idx, "index (rounded)": int(round(idx))}

    @classmethod
    def from_params(cls, params) -> Hill:
        """Create a Hill function from pre-determined parameters. Useful to recreate a Hill function"""
        instance = cls()
        instance.params = params
        return instance

    def gradient_at(self, x: float) -> float:
        """Return the gradient of the Hill function at a given x-value"""
        cxd = math.pow(self.params[2] / x, self.params[3])
        return (
            (self.params[1] - self.params[0])
            * self.params[3]
            * cxd
            / (math.pow(cxd + 1, 2) * x)
        )

    def x(self, y: float) -> float:
        """Return the x-value given a y-value"""
        return self.params[2] * math.pow(
            (y - self.params[0]) / (self.params[1] - y), 1 / self.params[3]
        )

    def y(self, x: float) -> float:
        """Return the y-value given an x-value."""
        return self.params[0] + (self.params[1] - self.params[0]) / (
            1 + (self.params[2] / x) ** self.params[3]
        )


def hill_func(x, a, b, c, d):  # Hill function
    """Calculates the Hill function at x.

    a : sigmoid low level
    b : sigmoid high level
    c : approximate inflection point
    d : slope of the sigmoid
    """
    return a + (b - a) / (1.0 + (c / x) ** d)



"""Module for classes that represent common geometric objects or patterns."""
from __future__ import annotations

import math
from itertools import zip_longest
from typing import Union, Optional, List, Iterable, Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Circle as mpl_Circle, Rectangle as mpl_Rectangle

from .utilities import is_iterable


def tan(degrees: float) -> float:
    """Calculate the tangent of the given degrees."""
    return math.tan(math.radians(degrees))


def cos(degrees: float) -> float:
    """Calculate the cosine of the given degrees."""
    return math.cos(math.radians(degrees))


def sin(degrees: float) -> float:
    """Calculate the sine of the given degrees."""
    return math.sin(math.radians(degrees))


class Point:
    """A geometric point with x, y, and z coordinates/attributes."""

    z: float
    y: float
    x: float
    _attr_list: List[str] = ["x", "y", "z", "idx", "value"]
    _coord_list: List[str] = ["x", "y", "z"]

    def __init__(
        self,
        x: Union[float, tuple, Point] = 0,
        y: float = 0,
        z: float = 0,
        idx: Optional[int] = None,
        value: Optional[float] = None,
        as_int: bool = False,
    ):
        """
        Parameters
        ----------
        x : number-like, Point, iterable
            x-coordinate or iterable type containing all coordinates. If iterable, values are assumed to be in order: (x,y,z).
        y : number-like, optional
            y-coordinate
        idx : int, optional
            Index of point. Useful for sequential coordinates; e.g. a point on a circle profile is sometimes easier to describe
            in terms of its index rather than x,y coords.
        value : number-like, optional
            value at point location (e.g. pixel value of an image)
        as_int : boolean
            If True, coordinates are converted to integers.
        """
        if isinstance(x, Point):
            for attr in self._attr_list:
                item = getattr(x, attr, None)
                setattr(self, attr, item)
        elif is_iterable(x):
            for attr, item in zip_longest(self._attr_list, x, fillvalue=0):
                setattr(self, attr, item)
        else:
            self.x = x
            self.y = y
            self.z = z
            self.idx = idx
            self.value = value

        if as_int:
            self.x = int(round(self.x))
            self.y = int(round(self.y))
            self.z = int(round(self.z))

    def distance_to(self, thing: Union["Point", "Circle"]) -> float:
        """Calculate the distance to the given point.

        Parameters
        ----------
        thing : Circle, Point, 2 element iterable
            The other thing to calculate distance to.
        """
        if isinstance(thing, Circle):
            return abs(
                np.sqrt((self.x - thing.center.x) ** 2 + (self.y - thing.center.y) ** 2)
                - thing.radius
            )
        p = Point(thing)
        return math.sqrt(
            (self.x - p.x) ** 2 + (self.y - p.y) ** 2 + (self.z - p.z) ** 2
        )

    def as_array(self, only_coords: bool = True) -> np.array:
        """Return the point as a numpy array."""
        if only_coords:
            return np.array([getattr(self, item) for item in self._coord_list])
        else:
            return np.array(
                [
                    getattr(self, item)
                    for item in self._attr_list
                    if (getattr(self, item) is not None)
                ]
            )

    def as_vector(self) -> "Vector":
        return Vector(x=self.x, y=self.y, z=self.z)

    def __repr__(self):
        return f"Point(x={self.x:3.2f}, y={self.y:3.2f}, z={self.z:3.2f})"

    def __eq__(self, other):
        # if all attrs equal, points considered equal
        return all(
            getattr(self, attr) == getattr(other, attr) for attr in self._attr_list
        )

    def __sub__(self, other):
        p = Point()
        for attr in self._attr_list:
            try:
                diff = getattr(self, attr) - getattr(other, attr)
            except TypeError:
                diff = None
            setattr(p, attr, diff)
        return p

    def __mul__(self, other):
        for attr in self._attr_list:
            try:
                self.__dict__[attr] *= other
            except TypeError:
                pass

    def __truediv__(self, other):
        for attr in self._attr_list:
            val = getattr(self, attr)
            # sometimes not all attrs are defined (like index or value and only x,y,z). Skip dividing those.
            if val is not None:
                setattr(self, attr, val / other)
        return self


class Circle:
    """A geometric circle with center Point, radius, and diameter."""

    center: Point
    radius: float

    def __init__(
        self, center_point: Union[Point, Iterable] = (0, 0), radius: float = 0
    ):
        """
        Parameters
        ----------
        center_point : Point, optional
            Center point of the wobble circle.
        radius : float, optional
            Radius of the wobble circle.
        """
        if center_point is None:
            center_point = Point()
        elif isinstance(center_point, Point) or is_iterable(center_point):
            center_point = Point(center_point)
        else:
            raise TypeError("Circle center must be of type Point or iterable")

        self.center = center_point
        self.radius = radius

    @property
    def diameter(self) -> float:
        """Get the diameter of the circle."""
        return self.radius * 2

    def plot2axes(
        self, axes: plt.Axes, edgecolor: str = "black", fill: bool = False
    ) -> None:
        """Plot the Circle on the axes.

        Parameters
        ----------
        axes : matplotlib.axes.Axes
            An MPL axes to plot to.
        edgecolor : str
            The color of the circle.
        fill : bool
            Whether to fill the circle with color or leave hollow.
        """
        axes.add_patch(
            mpl_Circle(
                (self.center.x, self.center.y),
                edgecolor=edgecolor,
                radius=self.radius,
                fill=fill,
            )
        )


class Vector:
    """A vector with x, y, and z coordinates."""

    x: float
    y: float
    z: float

    def __init__(self, x: float = 0, y: float = 0, z: float = 0):
        self.x = x
        self.y = y
        self.z = z

    def __repr__(self):
        return f"Vector(x={self.x:.2f}, y={self.y:.2f}, z={self.z:.2f})"

    def as_scalar(self) -> float:
        """Return the scalar equivalent of the vector."""
        return math.sqrt(self.x ** 2 + self.y ** 2 + self.z ** 2)

    def distance_to(self, thing: Union[Circle, Point]) -> float:
        """Calculate the distance to the given point.

        Parameters
        ----------
        thing : Circle, Point, 2 element iterable
            The other point to calculate distance to.
        """
        if isinstance(thing, Circle):
            return abs(
                np.sqrt((self.x - thing.center.x) ** 2 + (self.y - thing.center.y) ** 2)
                - thing.radius
            )
        else:
            p = Point(thing)
            return math.sqrt(
                (self.x - p.x) ** 2 + (self.y - p.y) ** 2 + (self.z - p.z) ** 2
            )

    def __sub__(self, other: "Vector") -> "Vector":
        new_x = self.x - other.x
        new_y = self.y - other.y
        new_z = self.z - other.z
        return Vector(x=new_x, y=new_y, z=new_z)

    def __add__(self, other: "Vector") -> "Vector":
        new_x = self.x + other.x
        new_y = self.y + other.y
        new_z = self.z + other.z
        return Vector(x=new_x, y=new_y, z=new_z)

    def __truediv__(self, other: float) -> "Vector":
        for attr in ("x", "y", "z"):
            val = getattr(self, attr)
            setattr(self, attr, val / other)
        return self


def vector_is_close(vector1: Vector, vector2: Vector, delta: float = 0.1) -> bool:
    """Determine if two vectors are with delta of each other; this is a simple coordinate comparison check."""
    for attr in ("x", "y", "z"):
        if np.isnan(getattr(vector1, attr)) and np.isnan(getattr(vector2, attr)):
            continue
        if (
            not getattr(vector2, attr) + delta
            >= getattr(vector1, attr)
            >= getattr(vector2, attr) - delta
        ):
            return False
    return True


class Line:
    """A line that is represented by two points or by m*x+b.

    Notes
    -----
    Calculations of slope, etc are from here:
    http://en.wikipedia.org/wiki/Linear_equation
    and here:
    http://www.mathsisfun.com/algebra/line-equation-2points.html
    """

    point1: Point
    point2: Point

    def __init__(
        self,
        point1: Union[Point, Tuple[float, float]],
        point2: Union[Point, Tuple[float, float]],
    ):
        """
        Parameters
        ----------
        point1 : Point
            One point of the line
        point2 : Point
            Second point along the line.
        """
        self.point1 = Point(point1)
        self.point2 = Point(point2)

    def __repr__(self):
        return (
            f"Line: p1:(x={self.point1.x:.1f}, y={self.point1.y:.1f}, z={self.point1.z:.1f}), "
            f"p2:(x={self.point2.x:.1f}, y={self.point2.y:.1f}, z={self.point2.z:.1f})"
        )

    @property
    def m(self) -> float:
        """Return the slope of the line.

        m = (y1 - y2)/(x1 - x2)

        From: http://www.purplemath.com/modules/slope.htm
        """
        return (self.point1.y - self.point2.y) / (self.point1.x - self.point2.x)

    @property
    def b(self) -> float:
        """Return the y-intercept of the line.

        b = y - m*x
        """
        return self.point1.y - (self.m * self.point1.x)

    def y(self, x) -> float:
        """Return y-value along line at position x."""
        return self.m * x + self.b

    def x(self, y) -> float:
        """Return x-value along line at position y."""
        return (y - self.b) / self.m

    @property
    def center(self) -> Point:
        """Return the center of the line as a Point."""
        mid_x = np.abs((self.point2.x - self.point1.x) / 2 + self.point1.x)
        mid_y = (self.point2.y - self.point1.y) / 2 + self.point1.y
        return Point(mid_x, mid_y)

    @property
    def length(self) -> float:
        """Return length of the line, if finite."""
        return self.point1.distance_to(self.point2)

    def distance_to(self, point: "Point") -> float:
        """Calculate the minimum distance from the line to a point.

        Equations are from here: http://mathworld.wolfram.com/Point-LineDistance2-Dimensional.html #14

        Parameters
        ----------
        point : Point, iterable
            The point to calculate distance to.
        """
        point = Point(point).as_array()
        lp1 = self.point1.as_array()
        lp2 = self.point2.as_array()
        numerator = np.sqrt(np.sum(np.power(np.cross((lp2 - lp1), (lp1 - point)), 2)))
        denominator = np.sqrt(np.sum(np.power(lp2 - lp1, 2)))
        return numerator / denominator

    def plot2axes(self, axes: plt.Axes, width: float = 1, color: str = "w") -> None:
        """Plot the line to an axes.

        Parameters
        ----------
        axes : matplotlib.axes.Axes
            An MPL axes to plot to.
        color : str
            The color of the line.
        """
        axes.plot(
            (self.point1.x, self.point2.x),
            (self.point1.y, self.point2.y),
            linewidth=width,
            color=color,
        )


class Rectangle:
    """A rectangle with width, height, center Point, top-left corner Point, and bottom-left corner Point."""

    width: Union[int, float]
    height: Union[int, float]
    _as_int: bool
    center: Point

    def __init__(
        self,
        width: float,
        height: float,
        center: Union[Point, Tuple],
        as_int: bool = False,
    ):
        """
        Parameters
        ----------
        width : number
            Width of the rectangle.
        height : number
            Height of the rectangle.
        center : Point, iterable, optional
            Center point of rectangle.
        as_int : bool
            If False (default), inputs are left as-is. If True, all inputs are converted to integers.
        """
        if as_int:
            self.width = int(np.round(width))
            self.height = int(np.round(height))
        else:
            self.width = width
            self.height = height
        self._as_int = as_int
        self.center = Point(center, as_int=as_int)

    @property
    def br_corner(self) -> Point:
        """The location of the bottom right corner."""
        return Point(
            self.center.x + self.width / 2,
            self.center.y - self.height / 2,
            as_int=self._as_int,
        )

    @property
    def bl_corner(self) -> Point:
        """The location of the bottom left corner."""
        return Point(
            self.center.x - self.width / 2,
            self.center.y - self.height / 2,
            as_int=self._as_int,
        )

    @property
    def tl_corner(self) -> Point:
        """The location of the top left corner."""
        return Point(
            self.center.x - self.width / 2,
            self.center.y + self.height / 2,
            as_int=self._as_int,
        )

    @property
    def tr_corner(self) -> Point:
        """The location of the top right corner."""
        return Point(
            self.center.x + self.width / 2,
            self.center.y + self.height / 2,
            as_int=self._as_int,
        )

    def plot2axes(
        self,
        axes: plt.Axes,
        edgecolor: str = "black",
        angle: float = 0.0,
        fill: bool = False,
        alpha: float = 1,
        facecolor: str = "g",
        label=None,
    ):
        """Plot the Rectangle to the axes.

        Parameters
        ----------
        axes : matplotlib.axes.Axes
            An MPL axes to plot to.
        edgecolor : str
            The color of the circle.
        angle : float
            Angle of the rectangle.
        fill : bool
            Whether to fill the rectangle with color or leave hollow.
        """
        axes.add_patch(
            mpl_Rectangle(
                (self.bl_corner.x, self.bl_corner.y),
                width=self.width,
                height=self.height,
                angle=angle,
                edgecolor=edgecolor,
                alpha=alpha,
                facecolor=facecolor,
                fill=fill,
                label=label,
            )
        )
